from builtins import max as _max
from builtins import min as _min


def max(*args, default=None):
    """Behaves like the built in max function but ignores any ``None`` values. e.g. ``max([1, None, 2]) == 2``.
    If the iterable is empty ``default`` will be returned.

    :param args: Single iterable or 1..n arguments
    :param default: Value that will be returned if the iterable is empty
    :return: max value
    """
    return _max(
        filter(lambda x: x is not None, args[0] if len(args) == 1 else args),
        default=default
    )


def min(*args, default=None):
    """Behaves like the built in min function but ignores any ``None`` values. e.g. ``min([1, None, 2]) == 1``.
    If the iterable is empty ``default`` will be returned.

    :param args: Single iterable or 1..n arguments
    :param default: Value that will be returned if the iterable is empty
    :return: min value
    """
    return _min(
        filter(lambda x: x is not None, args[0] if len(args) == 1 else args),
        default=default
    )



from typing import Union, TYPE_CHECKING, Optional, FrozenSet, Mapping

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from HABApp.openhab.items.commands import OnOffCommand, PercentCommand
from ..definitions import OnOffValue, PercentValue

if TYPE_CHECKING:
    Union = Union
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


class DimmerItem(OpenhabItem, OnOffCommand, PercentCommand):
    """DimmerItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Union[int, float] value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        try:
            return int(state)
        except ValueError:
            return float(state)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OnOffValue):
            new_value = 100 if new_value.on else 0
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        # Percent is 0 ... 100
        if isinstance(new_value, (int, float)):
            assert 0 <= new_value <= 100, new_value
        else:
            assert new_value is None, new_value

        return super().set_value(new_value)

    def __str__(self):
        return self.value

    def is_on(self) -> bool:
        """Test value against on-value"""
        return bool(self.value)

    def is_off(self) -> bool:
        """Test value against off-value"""
        return not bool(self.value)



from typing import Final, Iterable, List, Dict, TypeVar

from HABApp.core.const.const import PYTHON_311

if not PYTHON_311:
    from typing_extensions import Self
else:
    from typing import Self


class ValueFormatter:
    def __init__(self, value: str):
        self.value: Final = value

    def len(self):
        return len(self.value)

    def format(self, width: int) -> str:
        return f'{self.value:<{width}s}'


class EmptyFormatter(ValueFormatter):
    def __init__(self):
        super().__init__('')


TYPE_FORMATTER = TypeVar('TYPE_FORMATTER', bound=ValueFormatter)


class FormatterScope:
    def __init__(self, field_names: Iterable[str],
                 skip_alignment: Iterable[str] = (), min_width: Dict[str, int] = {}):
        self.lines: List[Dict[str, TYPE_FORMATTER]] = []
        self.keys: Final = tuple(field_names)

        # alignment options
        self._align_skip: Final = skip_alignment
        self._align_min_width: Final = min_width

        assert set(skip_alignment).issubset(self.keys)

    def add(self, obj: Dict[str, TYPE_FORMATTER]) -> Self:
        assert set(obj.keys()).issubset(self.keys)
        self.lines.append(obj)
        return self

    def get_indent_dict(self, ) -> Dict[str, int]:
        columns: Dict[str, List[TYPE_FORMATTER]] = {key: [] for key in self.keys}
        for line_dict in self.lines:
            for key in self.keys:
                formatter = line_dict.get(key)
                if formatter is None:
                    formatter = EmptyFormatter()
                columns[key].append(formatter)

        column_width = {key: max(map(lambda x: x.len(), column)) for key, column in columns.items()}

        for key, width in column_width.items():
            # indent to multiples of 4, if the entries are missing do not indent
            if width and key not in self._align_skip:
                add = width % 4
                if not add:
                    add = 4
                width += add

            # option to set a minimum width
            width = max(width, self._align_min_width.get(key, 0))

            column_width[key] = width

        return column_width

    def get_lines(self) -> List[str]:
        if not self.lines:
            return []

        column_width = self.get_indent_dict()

        ret_lines = []
        for line_dict in self.lines:    # type: Dict[str, TYPE_FORMATTER]
            line_vals = []
            for key, value_formatter in line_dict.items():
                width = column_width.get(key, 0)
                if width == 0:
                    continue
                line_vals.append(value_formatter.format(width))
            ret_lines.append(''.join(line_vals).rstrip())

        return ret_lines



from datetime import datetime

from HABApp.core.const.const import PYTHON_311
from HABApp.openhab.definitions import HSBValue, OnOffValue, OpenClosedValue, PercentValue, QuantityValue, RawValue, \
    UpDownValue
from HABApp.openhab.definitions.values import PointValue


def map_openhab_values(openhab_type: str, openhab_value: str):
    assert isinstance(openhab_type, str), type(openhab_type)
    assert isinstance(openhab_value, str), type(openhab_value)

    if openhab_type == 'UnDef' or openhab_value == 'NULL':
        return None

    if openhab_type == "Number":
        return int(openhab_value)

    if openhab_type == "Decimal":
        try:
            return int(openhab_value)
        except ValueError:
            return float(openhab_value)

    if openhab_type == "String":
        return openhab_value

    if openhab_type == "HSB":
        return HSBValue(openhab_value)

    if openhab_type == "DateTime":
        # see implementation im datetime_item.py
        if PYTHON_311:
            dt = datetime.fromisoformat(openhab_value)
        else:
            pos_dot = openhab_value.find('.')
            pos_plus = openhab_value.find('+')
            if pos_plus - pos_dot > 6:
                openhab_value = openhab_value[:pos_dot + 7] + openhab_value[pos_plus:]
            dt = datetime.strptime(openhab_value, '%Y-%m-%dT%H:%M:%S.%f%z')

        # all datetimes from openHAB have a timezone set, so we can't easily compare them
        # --> TypeError: can't compare offset-naive and offset-aware datetimes
        dt = dt.astimezone(tz=None)   # Changes datetime object so it uses system timezone
        dt = dt.replace(tzinfo=None)  # Removes timezone awareness
        return dt

    if openhab_type == 'OnOff':
        return OnOffValue(openhab_value)

    if openhab_type == 'OpenClosed':
        return OpenClosedValue(openhab_value)

    if openhab_type == 'UpDown':
        return UpDownValue(openhab_value)

    if openhab_type == 'Percent':
        return PercentValue(openhab_value)

    if openhab_type == 'Quantity':
        return QuantityValue(openhab_value)

    if openhab_type == 'Raw':
        return RawValue(openhab_value)

    if openhab_type == 'Point':
        return PointValue(openhab_value)

    return openhab_value



import asyncio
import logging
import sys
import typing

import HABApp
from HABApp.__cmd_args__ import parse_args, find_config_folder
from HABApp.__debug_info__ import print_debug_info
from HABApp.__splash_screen__ import show_screen


def main() -> typing.Union[int, str]:

    show_screen()

    # This has to be done before we create HABApp because of the possible sleep time
    args = parse_args()

    if HABApp.__cmd_args__.DO_DEBUG:
        print_debug_info()
        sys.exit(0)

    log = logging.getLogger('HABApp')

    try:
        cfg_folder = find_config_folder(args.config)

        # see if we have user code (e.g. for additional logging configuration or additional setup)
        try:
            import HABAppUser   # noqa: F401
        except ModuleNotFoundError:
            pass

        # Shutdown handler for graceful shutdown
        HABApp.runtime.shutdown.register_signal_handler()

        app = HABApp.runtime.Runtime()
        HABApp.core.const.loop.create_task(app.start(cfg_folder))
        HABApp.core.const.loop.run_forever()
    except Exception as e:
        for line in HABApp.core.lib.exceptions.format_exception(e):
            log.error(line)
            print(e)
        return str(e)
    finally:
        # Sleep to allow underlying connections of aiohttp to close
        # https://aiohttp.readthedocs.io/en/stable/client_advanced.html#graceful-shutdown
        HABApp.core.const.loop.run_until_complete(asyncio.sleep(1))

        asyncio.set_event_loop(None)
        HABApp.core.const.loop.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from typing import Generic, TypeVar, Literal, Union, Iterator, Tuple

from HABApp.core.const.const import PYTHON_310

if PYTHON_310:
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


T = TypeVar('T')

T_PRIO: TypeAlias = Union[Literal['first', 'last'], int]
T_ENTRY: TypeAlias = Tuple[T_PRIO, T]


def sort_func(obj: T_ENTRY):
    prio = {'first': 0, 'last': 2}
    key = obj[0]
    assert isinstance(key, int) or key in prio
    return prio.get(key, 1), key


class PriorityList(Generic[T]):
    def __init__(self):
        self._objs: list[T_ENTRY] = []

    def append(self, obj: T, priority: T_PRIO):
        for o in self._objs:
            assert o[0] != priority, priority
        self._objs.append((priority, obj))
        self._objs.sort(key=sort_func)

    def remove(self, obj: T):
        for i, (_, existing) in self._objs:
            if existing is obj:
                self._objs.pop(i)
                return None

    def __iter__(self) -> Iterator[T]:
        for p, o in self._objs:
            yield o

    def reversed(self) -> Iterator[T]:
        for p, o in reversed(self._objs):
            yield o

    def __repr__(self):
        return f'<{self.__class__.__name__} {[o for o in self]}>'



from typing import Optional

import HABApp
from HABApp.core.const.topics import TOPIC_ERRORS


class BenchBaseRule(HABApp.Rule):
    BENCH_TYPE: str

    def __init__(self):
        super().__init__()

        self.err_watcher = None
        self.errors = []

        self.prev_rule: Optional[BenchBaseRule] = None
        self.next_rule: Optional[BenchBaseRule] = None

    def link_rule(self, next_rule: 'BenchBaseRule'):
        assert self.next_rule is None
        assert next_rule.prev_rule is None

        self.next_rule = next_rule
        next_rule.prev_rule = self
        return next_rule

    def _err_event(self, event):
        self.errors.append(event)

    def do_bench_start(self):
        self.errors.clear()
        self.err_watcher = self.listen_event(TOPIC_ERRORS, self._err_event)

        self.run.at(1, self.do_bench_run)

    def do_bench_run(self):
        try:
            try:
                print('+' + '-' * 78 + '+')
                print(f'| {self.BENCH_TYPE:^76s} |')
                print('+' + '-' * 78 + '+')
                print('')

                self.set_up()
                self.run_bench()
            finally:
                self.tear_down()
        finally:
            self.run.at(1, self.do_bench_finished)

    def set_up(self):
        pass

    def tear_down(self):
        pass

    def run_bench(self):
        raise NotImplementedError()

    def do_bench_finished(self):
        self.err_watcher.cancel()

        if self.errors:
            count = len(self.errors)
            print(f'{count} error{"" if count == 1 else "s"} during Benchmark in {self.rule_name}!')
            for e in self.errors:
                print(f' - {type(e.exception)}: {e.exception}')

        if self.next_rule is None:
            HABApp.runtime.shutdown.request_shutdown()
        else:
            self.run.soon(self.next_rule.do_bench_start)



from __future__ import annotations

import asyncio
from time import monotonic

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.connection.handler.func_async import async_get_system_info


async def _start_level_reached() -> tuple[bool, None | int]:
    start_level_min = HABApp.CONFIG.openhab.general.min_start_level

    if (system_info := await async_get_system_info()) is None:
        return False, None

    start_level_is = system_info.start_level

    return start_level_is >= start_level_min, start_level_is


class WaitForStartlevelPlugin(BaseConnectionPlugin[OpenhabConnection]):

    async def on_connected(self, context: OpenhabContext, connection: OpenhabConnection):
        level_reached, level = await _start_level_reached()

        if level_reached:
            context.waited_for_openhab = False
            return None

        context.waited_for_openhab = False
        log = connection.log

        log.info('Waiting for openHAB startup to be complete')

        last_level: int = -100

        timeout_duration = 10 * 60
        timeout_start_at_level = 70
        timeout_timestamp = 0

        while not level_reached:
            await asyncio.sleep(1)

            level_reached, level = await _start_level_reached()

            # show start level change
            if last_level != level:
                if level is None:
                    log.debug('Start level: not received!')
                    level = -10
                else:
                    log.debug(f'Start level: {level}')

                # Wait but start eventually because sometimes we have a bad configured thing or an offline gateway
                # that prevents the start level from advancing
                # This is a safety net, so we properly start e.g. after a power outage
                # When starting manually one should fix the blocking thing
                if level >= timeout_start_at_level:
                    timeout_timestamp = monotonic()
                    log.debug('Starting start level timeout')

            # timeout is running
            if timeout_timestamp and monotonic() - timeout_timestamp > timeout_duration:
                log.warning(f'Starting even though openHAB is not ready yet (start level: {level})')
                break

            # update last level!
            last_level = level

        log.info('openHAB startup complete')



import logging
import re
import sys
import warnings
from pathlib import Path
from typing import Iterable, Union, Any, Optional, Tuple, Pattern, List, overload, Literal, TypeVar, Callable, Final

import HABApp
import HABApp.core
import HABApp.openhab
import HABApp.rule_manager
import HABApp.util
from HABApp.core.asyncio import create_task
from HABApp.core.const.const import PYTHON_310
from HABApp.core.const.hints import HINT_EVENT_CALLBACK
from HABApp.core.internals import HINT_EVENT_FILTER_OBJ, HINT_EVENT_BUS_LISTENER, ContextProvidingObj, \
    uses_post_event, EventFilterBase, uses_item_registry, ContextBoundEventBusListener
from HABApp.core.internals import wrap_func
from HABApp.core.items import BaseItem, HINT_ITEM_OBJ, HINT_TYPE_ITEM_OBJ, BaseValueItem
from HABApp.rule import interfaces
from HABApp.rule.scheduler import HABAppSchedulerView as _HABAppSchedulerView
from .interfaces import async_subprocess_exec
from .interfaces.rule_subprocess import build_exec_params, HINT_PYTHON_PATH, HINT_EXEC_ARGS, HINT_PROCESS_CB_SIMPLE, \
    HINT_PROCESS_CB_FULL
from .rule_hook import get_rule_hook as _get_rule_hook

if PYTHON_310:
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


log = logging.getLogger('HABApp.Rule')


# Func to log deprecation warnings
def send_warnings_to_log(message, category, filename, lineno, file=None, line=None):
    log.warning(f'{filename}:{lineno}: {category.__name__}:{message}')
    return


# Setup deprecation warnings
warnings.simplefilter('default')
warnings.showwarning = send_warnings_to_log


post_event = uses_post_event()
item_registry = uses_item_registry()


class Rule(ContextProvidingObj):

    def __init__(self):
        super().__init__(context=HABApp.rule_ctx.HABAppRuleContext(self))

        hook = _get_rule_hook()
        hook.register_rule(self)

        self.__runtime: HABApp.runtime.Runtime = hook.runtime
        assert isinstance(self.__runtime, HABApp.runtime.Runtime)

        # scheduler
        self.run: _HABAppSchedulerView = _HABAppSchedulerView(self._habapp_ctx)

        # suggest a rule name
        self.rule_name: str = hook.suggest_rule_name(self)

        # interfaces
        self.async_http = interfaces.http
        self.mqtt: Final = HABApp.mqtt.interface_sync
        self.oh: Final = HABApp.openhab.interface_sync
        self.openhab: Final = self.oh

    def on_rule_loaded(self):
        """Override this to implement logic that will be called when the rule and the file has been successfully loaded
        """

    def on_rule_removed(self):
        """Override this to implement logic that will be called when the rule has been unloaded.
        """

    def __repr__(self):
        # empty string, so we have a space if we have more than one entry
        parts = ['']

        # rule name if it is different from the class
        cls_name = self.__class__.__name__
        rule_name = str(self.rule_name)
        if cls_name != rule_name:
            parts.append(rule_name)

        # rule status
        if self._habapp_ctx is None:
            parts.append('(unloaded)')

        return f'<{cls_name}{" ".join(parts)}>'

    def post_event(self, name: Union[HINT_ITEM_OBJ, str], event: Any):
        """
        Post an event to the event bus

        :param name: name or item to post event to
        :param event: Event class to be used (must be class instance)
        :return:
        """
        assert isinstance(name, (str, BaseValueItem)), type(name)
        return post_event(
            name.name if isinstance(name, BaseValueItem) else name,
            event
        )

    def listen_event(self, name: Union[HINT_ITEM_OBJ, str],
                     callback: HINT_EVENT_CALLBACK,
                     event_filter: Optional[HINT_EVENT_FILTER_OBJ] = None
                     ) -> HINT_EVENT_BUS_LISTENER:
        """
        Register an event listener

        :param name: item or name to listen to
        :param callback: callback that accepts one parameter which will contain the event
        :param event_filter: Event filter. This is typically :class:`~HABApp.core.events.ValueUpdateEventFilter` or
            :class:`~HABApp.core.events.ValueChangeEventFilter` which will also trigger on changes/update from openhab
            or mqtt. Additionally it can be an instance of :class:`~HABApp.core.events.EventFilter` which additionally
            filters on the values of the event. It is also possible to group filters logically with, e.g.
            :class:`~HABApp.core.events.AndFilterGroup` and :class:`~HABApp.core.events.OrFilterGroup`
        """
        cb = wrap_func(callback, context=self._habapp_ctx)
        name = name.name if isinstance(name, BaseItem) else name

        if event_filter is None:
            event_filter = HABApp.core.events.NoEventFilter()
        if not isinstance(event_filter, EventFilterBase):
            raise ValueError(f'Argument event_filter must be an instance of event filter (is {event_filter})')

        listener = ContextBoundEventBusListener(name, cb, event_filter, parent_ctx=self._habapp_ctx)
        return self._habapp_ctx.add_event_listener(listener)

    @overload
    def execute_subprocess(self, callback: HINT_PROCESS_CB_SIMPLE, program: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                           additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                           raw_info: Literal[False], **kwargs):
        ...

    @overload
    def execute_subprocess(self, callback: HINT_PROCESS_CB_FULL, program: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                           additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                           raw_info: Literal[True], **kwargs):
        ...

    def execute_subprocess(self, callback, program: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                           additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                           raw_info: bool = False, **kwargs):
        """Run another program

        :param callback: Function that will be called when the process has finished.
                         First parameter takes a :class:`str` when :attr:`raw_info` is ``False`` (default) else
                         an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param program: python module (path to file) or python package
        :param args: arguments passed to the module or to package
        :param raw_info: ``False``: Return only the textual process output.
                         In case of failure (return code != 0) a log entry and an error event will be created.
                         This is the default and should be fine for almost all use cases.

                         ``True``: The callback will always be called with an
                         instance of :class:`~HABApp.rule.FinishedProcessInfo`.
        :param capture_output: Capture program output, set to ``False`` to only capture the return code
        :param additional_python_path: additional folders which will be added to the env variable ``PYTHONPATH``
        :param kwargs: Additional kwargs that will be passed to ``asyncio.create_subprocess_exec``
        :return:
        """

        cb = wrap_func(callback, context=self._habapp_ctx)

        call_args, call_kwargs = build_exec_params(
            program, *args, _capture_output=capture_output, _additional_python_path=additional_python_path, **kwargs
        )
        return create_task(
            async_subprocess_exec(
                cb.run, *call_args, raw_info=raw_info, calling_func=self.execute_python, **call_kwargs)
        )

    @overload
    def execute_python(self, callback: HINT_PROCESS_CB_SIMPLE, module_or_package: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                       additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                       raw_info: Literal[False], **kwargs):
        ...

    @overload
    def execute_python(self, callback: HINT_PROCESS_CB_FULL, module_or_package: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                       additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                       raw_info: Literal[True], **kwargs):
        ...

    def execute_python(self, callback, module_or_package: HINT_EXEC_ARGS, *args: HINT_EXEC_ARGS,
                       additional_python_path: HINT_PYTHON_PATH = None, capture_output: bool = True,
                       raw_info: bool = False, **kwargs):
        """Run a python module or package as a new process. The python environment that is used to run HABApp will be
        to run the module or package.

        :param callback: Function that will be called when the process has finished.
                         First parameter takes a :class:`str` when :attr:`raw_info` is ``False`` (default) else
                         an instance of :class:`~HABApp.rule.FinishedProcessInfo`
        :param module_or_package: python module (path to file) or python package (just the name)
        :param args: arguments passed to the module or to package
        :param raw_info: ``False``: Return only the textual process output.
                         In case of failure (return code != 0) a log entry and an error event will be created.
                         This is the default and should be fine for almost all use cases.

                         ``True``: The callback will always be called with an
                         instance of :class:`~HABApp.rule.FinishedProcessInfo`.
        :param capture_output: Capture program output, set to ``False`` to only capture the return code
        :param additional_python_path: additional folders which will be added to the env variable ``PYTHONPATH``
        :param kwargs: Additional kwargs that will be passed to ``asyncio.create_subprocess_exec``
        :return:
        """

        new_args = list(args)

        p = Path(module_or_package)
        if p.suffix.lower() == '.py':
            # if it's a relative path make it relative to the config
            if not p.is_absolute():
                p = (HABApp.CONFIG._file_path.parent / p).resolve()

            new_args.insert(0, p)

            # set parent folder as working directory for python script
            if 'cwd' not in kwargs:
                kwargs['cwd'] = p.parent
        else:
            new_args.insert(0, module_or_package)
            new_args.insert(0, '-m')

        cb = wrap_func(callback, context=self._habapp_ctx)
        call_args, call_kwargs = build_exec_params(
            sys.executable, *new_args,
            _capture_output=capture_output, _additional_python_path=additional_python_path,
            **kwargs
        )

        return create_task(
            async_subprocess_exec(
                cb.run, *call_args, raw_info=raw_info, calling_func=self.execute_python, **call_kwargs)
        )

    def get_rule(self, rule_name: str) -> 'Union[Rule, List[Rule]]':
        assert rule_name is None or isinstance(rule_name, str), type(rule_name)
        return self.__runtime.rule_manager.get_rule(rule_name)

    @staticmethod
    def get_items(type: Union[Tuple[HINT_TYPE_ITEM_OBJ, ...], HINT_TYPE_ITEM_OBJ] = None,
                  name: Union[str, Pattern[str]] = None,
                  tags: Union[str, Iterable[str]] = None,
                  groups: Union[str, Iterable[str]] = None,
                  metadata: Union[str, Pattern[str]] = None,
                  metadata_value: Union[str, Pattern[str]] = None,
                  ) -> Union[List[HINT_ITEM_OBJ], List[BaseItem]]:
        """Search the HABApp item registry and return the found items.

        :param type: item has to be an instance of this class
        :param name: str (will be compiled) or regex that is used to search the Name
        :param tags: item must have these tags (will return only instances of OpenhabItem)
        :param groups: item must be a member of these groups (will return only instances of OpenhabItem)
        :param metadata: str (will be compiled) or regex that is used to search the metadata (e.g. 'homekit')
        :param metadata_value: str (will be compiled) or regex that is used to search the metadata value
                               (e.g. 'TargetTemperature')
        :return: Items that match all the passed criteria
        """

        if name is not None and isinstance(name, str):
            name = re.compile(name, re.IGNORECASE)
        if metadata is not None and isinstance(metadata, str):
            metadata = re.compile(metadata, re.IGNORECASE)
        if metadata_value is not None and isinstance(metadata_value, str):
            metadata_value = re.compile(metadata_value, re.IGNORECASE)

        _tags, _groups = None, None
        if tags is not None:
            _tags = set(tags) if not isinstance(tags, str) else {tags}
        if groups is not None:
            _groups = set(groups) if not isinstance(groups, str) else {groups}

        OpenhabItem = HABApp.openhab.items.OpenhabItem
        if _tags or _groups or metadata or metadata_value:
            if type is None:
                type = OpenhabItem
            if not issubclass(type, OpenhabItem):
                raise ValueError('Searching for tags, groups and metadata only works for OpenhabItem or its subclasses')

        ret = []
        for item in item_registry.get_items():  # type: HABApp.core.items.BaseItem
            if type is not None and not isinstance(item, type):
                continue

            if name is not None and not name.search(item.name):
                continue

            if _tags is not None and not _tags.issubset(item.tags):
                continue

            if _groups is not None and not _groups.issubset(item.groups):
                continue

            if metadata is not None and not any(map(metadata.search, item.metadata)):
                continue

            if metadata_value is not None and not any(
                    map(metadata_value.search, map(lambda x: x[0], item.metadata.values()))):
                continue

            ret.append(item)
        return ret


PSPEC_RULE = ParamSpec('PSPEC_RULE')
TYPE_RULE = TypeVar('TYPE_RULE', bound=Rule)


def create_rule(f: Callable[PSPEC_RULE, TYPE_RULE], *args: PSPEC_RULE.args, **kwargs: PSPEC_RULE.kwargs) -> TYPE_RULE:
    try:
        _get_rule_hook()
    except RuntimeError:
        return None

    return f(*args, **kwargs)



import logging
from typing import Tuple, Any, Optional

from aiomqtt import Message

from HABApp.core.const.json import load_json
from HABApp.core.const.log import TOPIC_EVENTS
from HABApp.core.wrapper import process_exception

log = logging.getLogger(f'{TOPIC_EVENTS}.mqtt')


def get_msg_payload(msg: Message) -> Tuple[Optional[str], Any]:
    try:
        topic = msg.topic.value
        raw = msg.payload

        try:
            val = raw.decode("utf-8")
        except UnicodeDecodeError:
            # Payload ist a byte stream
            if log.isEnabledFor(logging.DEBUG):
                log._log(logging.DEBUG, f'{topic} ({msg.qos}): {raw[:20]}...', [])
            return topic, raw

        if log.isEnabledFor(logging.DEBUG):
            log._log(logging.DEBUG, f'{topic} ({msg.qos}): {val}', [])

        # None
        if val == 'none' or val == 'None':
            return topic, None

        # bool
        if val == 'true' or val == 'True':
            return topic, True
        if val == 'false' or val == 'False':
            return topic, False

        # int
        if val.isdecimal():
            return topic, int(val)

        # json list/dict
        if val.startswith('{') and val.endswith('}') or val.startswith('[') and val.endswith(']'):
            try:
                return topic, load_json(val)
            except ValueError:
                return topic, val

        # float or str
        try:
            return topic, float(val)
        except ValueError:
            return topic, val
    except Exception as e:
        process_exception(get_msg_payload, e, logger=log)
        return None, None



import logging
import threading
from typing import Dict
from typing import Tuple, Union, TypeVar

from HABApp.core.errors import ItemNotFoundException, ItemAlreadyExistsError
from HABApp.core.internals.item_registry import ItemRegistryItem


_HINT_ITEM_OBJ = TypeVar('_HINT_ITEM_OBJ', bound=ItemRegistryItem)

log = logging.getLogger('HABApp.Items')


class ItemRegistry:
    def __init__(self):
        self._lock = threading.Lock()
        self._items: Dict[str, ItemRegistryItem] = {}

    def item_exists(self, name: Union[str, ItemRegistryItem]) -> bool:
        if not isinstance(name, str):
            name = name.name
        return name in self._items

    def get_item(self, name: str) -> ItemRegistryItem:
        try:
            return self._items[name]
        except KeyError:
            raise ItemNotFoundException(name) from None

    def get_items(self) -> Tuple[ItemRegistryItem, ...]:
        return tuple(self._items.values())

    def get_item_names(self) -> Tuple[str, ...]:
        return tuple(self._items.keys())

    def add_item(self, item: _HINT_ITEM_OBJ) -> _HINT_ITEM_OBJ:
        assert isinstance(item, ItemRegistryItem)
        name = item.name

        with self._lock:
            existing = self._items.get(name)
            if existing is not None:
                # adding the same item multiple times will not cause an exception
                if existing is item:
                    return item

                # adding a new item with the same name raises an exception
                raise ItemAlreadyExistsError(name)

            self._items[name] = item

        log.debug(f'Added {name} ({item.__class__.__name__})')
        item._on_item_added()
        return item

    def pop_item(self, name: Union[str, _HINT_ITEM_OBJ]) -> _HINT_ITEM_OBJ:
        if not isinstance(name, str):
            name = name.name

        with self._lock:
            try:
                item = self._items.pop(name)
            except KeyError:
                raise ItemNotFoundException(name) from None

        log.debug(f'Removed {name} ({item.__class__.__name__})')
        item._on_item_removed()
        return item



class HABAppException(Exception):
    pass


class ProxyObjHasNotBeenReplacedError(HABAppException):
    def __init__(self, obj) -> None:
        super().__init__(f'{obj} has not been replaced on startup!')


class ItemNotFoundException(HABAppException):
    def __init__(self, name: str):
        super().__init__(f'Item {name} does not exist!')
        self.name: str = name


class ItemAlreadyExistsError(HABAppException):
    def __init__(self, name: str):
        super().__init__(f'Item {name} does already exist and can not be added again!')
        self.name: str = name


class ContextNotFoundError(HABAppException):
    pass


class ContextNotSetError(HABAppException):
    pass


class ContextBoundObjectIsAlreadyLinkedError(HABAppException):
    pass


class ContextBoundObjectIsAlreadyUnlinkedError(HABAppException):
    pass



import logging
import typing

import HABApp
from HABApp.core.asyncio import run_func_from_async
from HABApp.core.events import ItemNoChangeEvent, ItemNoUpdateEvent, EventFilter
from HABApp.core.lib import PendingFuture
from HABApp.core.const.hints import HINT_EVENT_CALLBACK
from HABApp.core.internals import uses_post_event, get_current_context, AutoContextBoundObj, wrap_func
from HABApp.core.internals import ContextBoundEventBusListener

log = logging.getLogger('HABApp')

post_event = uses_post_event()


class BaseWatch(AutoContextBoundObj):
    EVENT: typing.Union[typing.Type[ItemNoUpdateEvent], typing.Type[ItemNoChangeEvent]]

    def __init__(self, name: str, secs: typing.Union[int, float]):
        super().__init__()
        self.fut = PendingFuture(self._post_event, secs)
        self.name: str = name

    async def _post_event(self):
        post_event(self.name, self.EVENT(self.name, self.fut.secs))

    def __cancel_watch(self):
        self.fut.cancel()
        log.debug(f'Canceled {self.__class__.__name__} ({self.fut.secs}s) for {self.name}')

    def cancel(self):
        """Cancel the item watch"""
        self._ctx_unlink()
        run_func_from_async(self.__cancel_watch)

    def listen_event(self, callback: HINT_EVENT_CALLBACK) -> 'HABApp.core.base.HINT_EVENT_BUS_LISTENER':
        """Listen to (only) the event that is emitted by this watcher"""
        context = get_current_context()
        return context.add_event_listener(
            ContextBoundEventBusListener(
                self.name,
                wrap_func(callback, context=context),
                EventFilter(self.EVENT, seconds=self.fut.secs),
                parent_ctx=context
            )
        )


class ItemNoUpdateWatch(BaseWatch):
    EVENT = ItemNoUpdateEvent


class ItemNoChangeWatch(BaseWatch):
    EVENT = ItemNoChangeEvent



import random
import time
from collections import deque
from threading import Lock

import HABApp
from HABApp.core.events import ValueUpdateEvent
from .bench_base import BenchBaseRule
from .bench_times import BenchContainer, BenchTime

LOCK = Lock()


class HABAppBenchRule(BenchBaseRule):
    BENCH_TYPE = 'HABApp'

    def __init__(self):
        super().__init__()

        self.name_list = [f'BenchItem{k}' for k in range(300)]

        self.time_sent = 0.0
        self.bench_started = 0.0
        self.bench_times_container = BenchContainer()
        self.bench_times: BenchTime = None

        self.name = ''
        self.values = deque()

    def cleanup(self):
        for n in self.name_list:
            if HABApp.core.Items.item_exists(n):
                HABApp.core.Items.pop_item(n)

    def set_up(self):
        self.cleanup()

    def tear_down(self):
        pass
        self.cleanup()

    def run_bench(self):
        # These are the benchmarks
        self.bench_rtt_time()

    def bench_rtt_time(self):
        print('Bench events ', end='')
        self.bench_times_container = BenchContainer()

        self.run_rtt('rtt idle')
        self.run_rtt('async rtt idle', do_async=True)

        # self.start_load()
        # self.run_rtt('rtt load (+10x)')
        # self.run_rtt('async rtt load (+10x)', do_async=True)
        # self.stop_load()

        print(' done!\n')
        time.sleep(0.1)
        self.bench_times_container.show()

    def run_rtt(self, test_name, do_async=False):
        self.name = self.name_list[0]
        self.openhab.create_item('String', self.name, label='MyLabel')

        for i in range(50_000):
            self.values.append(random.randint(0, 99999999))

        listener = self.listen_event(
            self.name, self.post_next_event_val if not do_async else self.a_post_next_event_val
        )

        self.bench_times = self.bench_times_container.create(test_name)

        self.time_sent = time.time()
        self.post_event(self.name, self.values[0])

        self.run.soon(LOCK.acquire)
        time.sleep(1)
        LOCK.acquire(True, 5)

        listener.cancel()
        if LOCK.locked():
            LOCK.release()

        print('.', end='')

    def post_next_event_val(self, value):
        if value != self.values[0]:
            return None

        self.bench_times.times.append(time.time() - self.time_sent)

        # No items left -> stop benchmark
        try:
            self.values.popleft()
        except IndexError:
            LOCK.release()
            return None

        if not self.values:
            LOCK.release()
            return None

        self.time_sent = time.time()
        self.post_event(self.name, self.values[0])

    async def a_post_next_event_val(self, event: ValueUpdateEvent):
        self.post_next_event_val(event)



from asyncio import run_coroutine_threadsafe

from HABApp.core.const import loop
from HABApp.core.asyncio import async_context
from eascheduler.jobs.job_base import ScheduledJobBase
from eascheduler.schedulers import AsyncScheduler


class HABAppScheduler(AsyncScheduler):
    def __init__(self):
        super().__init__()

        # we start paused, so we execute stuff only if we can load the whole file properly
        self.pause()

    async def _run_next(self):
        async_context.set('HABAppScheduler')
        return await super()._run_next()

    async def __add_job(self, job: ScheduledJobBase):
        super().add_job(job)

    async def __remove_job(self, job: ScheduledJobBase):
        super().remove_job(job)

    async def __cancel_all(self):
        super().cancel_all()

    async def __pause(self):
        super().pause()

    async def __resume(self):
        super().resume()

    def pause(self):
        if async_context.get(None) is None:
            run_coroutine_threadsafe(self.__pause(), loop).result()
        else:
            super().pause()
        return None

    def resume(self):
        if async_context.get(None) is None:
            run_coroutine_threadsafe(self.__resume(), loop).result()
        else:
            super().resume()
        return None

    def add_job(self, job: ScheduledJobBase):
        if async_context.get(None) is None:
            run_coroutine_threadsafe(self.__add_job(job), loop).result()
        else:
            super().add_job(job)
        return None

    def remove_job(self, job: ScheduledJobBase):
        if async_context.get(None) is None:
            run_coroutine_threadsafe(self.__remove_job(job), loop).result()
        else:
            super().remove_job(job)
        return None

    def cancel_all(self):
        if async_context.get(None) is None:
            run_coroutine_threadsafe(self.__cancel_all(), loop).result()
        else:
            super().cancel_all()
        return None



import asyncio
import os
import sys


# we can have subprocesses (https://docs.python.org/3/library/asyncio-platforms.html#subprocess-support-on-windows)
# or mqtt support (https://github.com/sbtinstruments/aiomqtt#note-for-windows-users)
# but not both. For testing, it makes sense to use mqtt support as a default
if (sys.platform.lower() == "win32" or os.name.lower() == "nt") and os.environ.get('HABAPP_NO_MQTT') is None:
    from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy
    set_event_loop_policy(WindowsSelectorEventLoopPolicy())

loop = asyncio.new_event_loop()
asyncio.set_event_loop(loop)

loop.set_debug(True)
loop.slow_callback_duration = 0.02



from typing import TYPE_CHECKING, Optional, FrozenSet, Mapping, Union

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from HABApp.openhab.items.commands import UpDownCommand, PercentCommand
from ..definitions import UpDownValue, PercentValue

if TYPE_CHECKING:
    Union = Union
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


class RollershutterItem(OpenhabItem, UpDownCommand, PercentCommand):
    """RollershutterItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Union[int, float] value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        try:
            return int(state)
        except ValueError:
            return float(state)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, UpDownValue):
            new_value = 0 if new_value.up else 100
        elif isinstance(new_value, PercentValue):
            new_value = new_value.value

        assert isinstance(new_value, (int, float)) or new_value is None, new_value
        return super().set_value(new_value)

    def is_up(self) -> bool:
        return self.value <= 0

    def is_down(self) -> bool:
        return self.value >= 100

    def __str__(self):
        return self.value



# Version scheme:
# YEAR.MONTH.COUNTER
# YEAR, MONTH are two digits and zero-padded
# COUNTER resets every month and starts with 0
# For example:
# - 23.09.0
# - 23.09.1
# - 23.10.0
#
# Development versions contain the DEV-COUNTER postfix:
# - 23.09.0.DEV-1

__version__ = '23.09.1'



import asyncio
import functools
import logging
import typing
from logging import Logger
# noinspection PyProtectedMember
from sys import _getframe as sys_get_frame
from typing import Union, Callable

from HABApp.core.const.topics import TOPIC_ERRORS as TOPIC_ERRORS
from HABApp.core.const.topics import TOPIC_WARNINGS as TOPIC_WARNINGS
from HABApp.core.events.habapp_events import HABAppException
from HABApp.core.internals import uses_post_event
from HABApp.core.lib import format_exception

log = logging.getLogger('HABApp')

post_event = uses_post_event()


def process_exception(func: Union[Callable, str], e: Exception,
                      do_print=False, logger: logging.Logger = log):
    lines = format_exception(e)

    func_name = func if isinstance(func, str) else func.__name__

    # log exception, since it is unexpected we push it to stdout, too
    if do_print:
        print(f'Error {e} in {func_name}:')
    logger.error(f'Error {e} in {func_name}:')
    for line in lines:
        if do_print:
            print(line)
        logger.error(line)

    # send Error to internal event bus, so we can reprocess it and notify the user
    post_event(TOPIC_ERRORS, HABAppException(func_name=func_name, exception=e, traceback='\n'.join(lines)))


def log_exception(func):
    # return async wrapper
    if asyncio.iscoroutinefunction(func) or asyncio.iscoroutine(func):
        @functools.wraps(func)
        async def a(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                process_exception(func, e, do_print=True)
                # re raise exception, since this is something we didn't anticipate
                raise

        return a

    @functools.wraps(func)
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            process_exception(func, e, do_print=True)
            # re raise exception, since this is something we didn't anticipate
            raise

    return f


def ignore_exception(func):
    # return async wrapper
    if asyncio.iscoroutinefunction(func) or asyncio.iscoroutine(func):
        @functools.wraps(func)
        async def a(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                process_exception(func, e)
                return None

        return a

    @functools.wraps(func)
    def f(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            process_exception(func, e)
            return None
    return f


class ExceptionToHABApp:
    def __init__(self, logger: typing.Optional[Logger] = None, log_level: int = logging.ERROR,
                 ignore_exception: bool = True):
        self.log: typing.Optional[Logger] = logger
        self.log_level = log_level
        self.ignore_exception: bool = ignore_exception

        self.raised_exception = False

        self.proc_tb: typing.Optional[typing.Callable[[list], list]] = None

    def __enter__(self):
        self.raised_exception = False

    def __exit__(self, exc_type, exc_val, exc_tb):
        # no exception -> we exit gracefully
        if exc_type is None and exc_val is None:
            return True

        self.raised_exception = True

        tb = format_exception((exc_type, exc_val, exc_tb))

        # possibility to reprocess tb
        if self.proc_tb is not None:
            tb = self.proc_tb(tb)

        # try to get the parent function name
        try:
            f_name = sys_get_frame().f_back.f_code.co_name
        except Exception:
            f_name = 'Exception while getting the function name!'

        # log error
        if self.log is not None:
            self.log.log(self.log_level, f'Error "{exc_val}" in {f_name}:')
            for line in tb:
                self.log.log(self.log_level, line)

        # send Error to internal event bus so we can reprocess it and notify the user
        post_event(
            TOPIC_WARNINGS if self.log_level == logging.WARNING else TOPIC_ERRORS,
            HABAppException(func_name=f_name, exception=exc_val, traceback='\n'.join(tb))
        )
        return self.ignore_exception



import importlib
import sys
from typing import List, Dict

from HABApp.__debug_info__ import print_debug_info


def get_dependencies() -> List[str]:
    return [
        'aiohttp-sse-client',
        'aiohttp',
        'aiomqtt',
        'bidict',
        'colorama',
        'eascheduler',
        'easyconfig',
        'pydantic',
        'stack_data',
        'voluptuous',
        'watchdog',
        'ujson',
        'immutables',
        'javaproperties',
        'msgspec',
        'pendulum',

        'typing-extensions',
    ]


def check_dependency_packages():
    """Imports all dependencies and reports failures"""

    missing: Dict[str, ModuleNotFoundError] = {}

    # Package aliases (if the import name differs from the package name)
    alias = {
        'aiohttp-sse-client': 'aiohttp_sse_client',
        'paho-mqtt': 'paho.mqtt',
        'typing-extensions': 'typing_extensions',
    }

    for name in get_dependencies():
        try:
            importlib.import_module(alias.get(name, name))
        except ModuleNotFoundError as e:
            missing[name] = e

    if not missing:
        return None

    print_debug_info()
    print()

    print(f'Error: {len(missing)} package{"s are" if len(missing) != 1 else " is"} missing:')
    for name, err in missing.items():
        print(f' - {name}: {err}')

    sys.exit(100)



import logging
from typing import Any, Dict, FrozenSet, Optional

from immutables import Map

import HABApp
from HABApp.core.wrapper import process_exception
from HABApp.openhab.definitions.values import QuantityValue
from HABApp.openhab.items import ColorItem, ContactItem, DatetimeItem, DimmerItem, GroupItem, ImageItem, LocationItem, \
    NumberItem, PlayerItem, RollershutterItem, StringItem, SwitchItem, CallItem
from HABApp.openhab.items.base_item import HINT_TYPE_OPENHAB_ITEM
from HABApp.openhab.items.base_item import MetaData

log = logging.getLogger('HABApp.openhab.items')


_items: Dict[str, HINT_TYPE_OPENHAB_ITEM] = {
    'String': StringItem,
    'Number': NumberItem,
    'Switch': SwitchItem,
    'Contact': ContactItem,
    'Rollershutter': RollershutterItem,
    'Dimmer': DimmerItem,
    'DateTime': DatetimeItem,
    'Color': ColorItem,
    'Image': ImageItem,
    'Group': GroupItem,
    'Player': PlayerItem,
    'Location': LocationItem,
    'Call': CallItem,
}


def map_item(name: str, type: str, value: Optional[str],
             label: Optional[str], tags: FrozenSet[str],
             groups: FrozenSet[str], metadata: Optional[Dict[str, Dict[str, Any]]]) -> \
        Optional['HABApp.openhab.items.OpenhabItem']:
    try:
        assert isinstance(type, str)
        assert value is None or isinstance(value, str)

        # map Metadata
        if metadata is not None:
            meta = Map({k: MetaData(v['value'], Map(v.get('config', {}))) for k, v in metadata.items()})
        else:
            meta = Map()

        # Quantity types are like this: Number:Temperature and have a unit set: "12.3 °C".
        # We have to remove the dimension from the type and remove the unit from the value
        if ':' in type:
            type, dimension = type.split(':')
            # if the item is not initialized its None and has no dimension
            if value is not None:
                value, _ = QuantityValue.split_unit(value)

            # Show warning
            # https://github.com/spacemanspiff2007/HABApp/issues/383
            if 'unit' not in meta:
                log.warning(f'Item {name:s} is a UoM item but "unit" is not found in item metadata')

        cls = _items.get(type)
        if cls is not None:
            return cls.from_oh(name, value, label=label, tags=tags, groups=groups, metadata=meta)

        raise ValueError(f'Unknown openHAB type: {type} for {name}')

    except Exception as e:
        process_exception('map_items', e, logger=log)
        return None



from __future__ import annotations

import re
from dataclasses import dataclass
from inspect import signature, iscoroutinefunction, getmembers
from typing import Awaitable, Callable, Any, TYPE_CHECKING

from ._definitions import ConnectionStatus

if TYPE_CHECKING:
    from .base_connection import BaseConnection
    from .base_plugin import BaseConnectionPlugin


def get_plugin_callbacks(obj: BaseConnectionPlugin) -> list[tuple[ConnectionStatus, PluginCallbackHandler]]:
    name_to_status = {obj.lower(): obj for obj in ConnectionStatus}
    name_regex = re.compile(f'on_({"|".join(name_to_status)})')

    ret = []
    for m_name, member in getmembers(obj, predicate=lambda x: callable(x)):
        if not m_name.lower().startswith('on_'):
            continue

        if m_name in ('on_application_shutdown', ):
            continue

        if (m := name_regex.fullmatch(m_name)) is None:
            raise ValueError(f'Invalid name: {m_name} in {obj.plugin_name}')

        status = name_to_status[m.group(1)]
        cb = PluginCallbackHandler.create(obj, member)

        ret.append((status, cb))

    return ret


@dataclass
class PluginCallbackHandler:
    plugin: BaseConnectionPlugin
    coro: Callable[[...], Awaitable]
    kwargs: tuple[str, ...]

    async def run(self, connection: BaseConnection, context: Any):
        kwargs = {}
        if self.kwargs:
            if 'connection' in self.kwargs:
                kwargs['connection'] = connection
            if 'context' in self.kwargs:
                kwargs['context'] = context

        return await self.coro(**kwargs)

    @staticmethod
    def _get_coro_kwargs(plugin: BaseConnectionPlugin, coro: Callable[[...], Awaitable]):
        if not iscoroutinefunction(coro):
            raise ValueError(f'Coroutine function expected for {plugin.plugin_name}.{coro.__name__}')

        sig = signature(coro)

        kwargs = []
        for name in sig.parameters:
            if name in ('connection', 'context'):
                kwargs.append(name)
            else:
                raise ValueError(f'Invalid parameter name "{name:s}" for {plugin.plugin_name}.{coro.__name__}')
        return tuple(kwargs)

    @classmethod
    def create(cls, plugin: BaseConnectionPlugin, coro: Callable[[...], Awaitable]):
        return cls(plugin, coro, cls._get_coro_kwargs(plugin, coro))



from typing import Any, Mapping

from immutables import Map
from pendulum import UTC
from pendulum import now as pd_now

from HABApp.core.items import BaseItem
from HABApp.openhab.definitions import ThingStatusEnum, ThingStatusDetailEnum
from HABApp.openhab.definitions.things import THING_STATUS_DEFAULT, THING_STATUS_DETAIL_DEFAULT
from HABApp.openhab.events import ThingConfigStatusInfoEvent, ThingStatusInfoEvent, ThingUpdatedEvent
from HABApp.openhab.interface_sync import set_thing_enabled


class Thing(BaseItem):
    """Base class for Things

    :ivar ThingStatusEnum status: Status of the thing (e.g. OFFLINE, ONLINE, ...)
    :ivar ThingStatusDetailEnum status_detail: Additional detail for the status
    :ivar str status_description: Additional description for the status
    :ivar str label: Thing label
    :ivar str location: Thing location
    :ivar Mapping[str, Any] configuration: Thing configuration
    :ivar Mapping[str, Any] properties: Thing properties
    """
    def __init__(self, name: str):
        super().__init__(name)

        self.status: ThingStatusEnum = THING_STATUS_DEFAULT
        self.status_detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT
        self.status_description: str = ''

        self.label: str = ''
        self.location: str = ''

        self.configuration: Mapping[str, Any] = Map()
        self.properties: Mapping[str, Any] = Map()

    @property
    def is_enabled(self) -> bool:
        # https://github.com/openhab/openhab-core/issues/3055
        return self.status_detail != 'DISABLED'

    def __update_timestamps(self, changed: bool):
        _now = pd_now(UTC)
        self._last_update.set(_now)
        if changed:
            self._last_change.set(_now)
        return None

    def process_event(self, event):

        if isinstance(event, ThingStatusInfoEvent):
            old_status = self.status
            old_detail = self.status_detail
            old_description = self.status_description

            self.status = new_status = event.status
            self.status_detail = new_detail = event.detail
            self.status_description = new_description = event.description

            self.__update_timestamps(
                old_status != new_status or old_detail != new_detail or old_description != new_description
            )
        elif isinstance(event, ThingUpdatedEvent):
            old_label = self.label
            old_configuration = self.configuration
            old_properties = self.properties
            old_location = self.location

            self.label = new_label = event.label
            self.location = new_location = event.location
            self.configuration = new_configuration = Map(event.configuration)
            self.properties = new_properties = Map(event.properties)

            self.__update_timestamps(
                old_label != new_label or old_location != new_location or   # noqa: W504
                old_configuration != new_configuration or old_properties != new_properties
            )
        elif isinstance(event, ThingConfigStatusInfoEvent):
            pass

        return None

    def set_enabled(self, enable: bool = True):
        """Enable/disable the thing

        :param enable: True to enable, False to disable the thing
        :return:
        """
        return set_thing_enabled(self.name, enable)



from datetime import timedelta
from time import time
from typing import Union, Optional

from HABApp.core.internals import wrap_func, AutoContextBoundObj


VAL_TYPE = Union[int, float]


class FadeWorker(AutoContextBoundObj):

    def __init__(self, parent: 'Fade', interval: float):
        super().__init__()
        self.parent: 'Fade' = parent
        self.scheduler = self._parent_ctx.rule.run.every(None, interval, self.parent._scheduled_worker)

    def cancel(self):
        self._ctx_unlink()
        self.scheduler.cancel()
        self.scheduler = None

        self.parent._fade_worker = None
        self.parent = None


MIN_STEP_TIME = 0.2


class Fade:
    """Helper to easily fade values up/down

    :ivar min_value: minimum valid value for the fade value
    :ivar max_value: maximum valid value for the fade value
    :ivar callback: Function with one argument that will be automatically called with the new values when the scheduled
                    fade runs
    """
    def __init__(self, callback=None, min_value: VAL_TYPE = 0, max_value: VAL_TYPE = 100):
        self.min_value = min_value
        self.max_value = max_value

        self._fade_start_time = 0
        self._fade_start_value = 0
        self._fade_stop_value = 0
        self._step_duration = 0
        self._fade_factor = 0
        self._fade_finished = True

        self._fade_worker: Optional[FadeWorker] = None
        self.__callback = wrap_func(callback) if callback is not None else None

        self.value = 0

    def setup(self, start_value: VAL_TYPE, stop_value: VAL_TYPE, duration: Union[int, float, timedelta],
              min_step_duration: float = MIN_STEP_TIME, now: Optional[float] = None) -> 'Fade':
        """Calculates everything that is needed to fade a value

        :param start_value: Start value
        :param stop_value:  Stop value
        :param duration: How long shall the fade take
        :param min_step_duration: minimum step duration (min 0.2 secs)
        :param now: time.time() timestamp to sync multiple fades together
        """
        if start_value < self.min_value or start_value > self.max_value:
            raise ValueError('Start value is out of range')
        if stop_value < self.min_value or stop_value > self.max_value:
            raise ValueError('Stop value is out of range')

        if isinstance(duration, timedelta):
            duration = duration.total_seconds()
        assert isinstance(duration, (int, float)) and duration >= 1

        diff = stop_value - start_value
        if not diff:
            raise ValueError('Start value must be different than stop value')

        # If we have a running fade we have to cancel it before changing the used values
        self.stop_fade()

        self._fade_start_value = start_value
        self._fade_stop_value = stop_value

        self._fade_factor = diff / duration
        self._step_duration = max(MIN_STEP_TIME, min_step_duration, 1 / abs(self._fade_factor))

        self._fade_start_time = time() if now is None else now
        self._fade_finished = False
        return self

    def get_value(self, now: Optional[float] = None) -> float:
        """Returns the current value. If the fade is finished it will always return the stop value.

        :param now: time.time() timestamp for which the value shall be returned. Can be used to sync multiple fades
                    together. Not required.
        :return: current value
        """
        if self._fade_finished:
            return self.value

        if now is None:
            now = time()

        assert now > self._fade_start_time
        dur = now - self._fade_start_time
        value = self._fade_start_value + self._fade_factor * dur

        if self._fade_factor < 0:
            if value <= self._fade_stop_value:
                self._fade_finished = True
                value = max(value, self._fade_stop_value)
        else:
            if value >= self._fade_stop_value:
                self._fade_finished = True
                value = min(value, self._fade_stop_value)

        self.value = value = round(value, 2)
        return value

    @property
    def is_finished(self) -> bool:
        """True if the fade is finished"""
        return self._fade_finished

    async def _scheduled_worker(self):
        self.get_value()
        if self._fade_finished:
            self.stop_fade()

        if self.__callback is not None:
            self.__callback.run(self.value)

    def schedule_fade(self) -> 'Fade':
        """Automatically run the fade with the Scheduler. The callback can be used to set the current fade value
        e.g. on an item. Calling this on a running fade will restart the fade
        """
        if self._fade_worker is not None:
            self._fade_worker.cancel()  # This will also set this variable to None, so we don't have to do it

        self._fade_worker = FadeWorker(self, self._step_duration)
        return self

    def stop_fade(self):
        """Stop the scheduled fade. This can be called multiple times without error"""
        if self._fade_worker is None:
            return None
        self._fade_worker.cancel()



from easyconfig import AppBaseModel
from HABApp.config.models.location import LocationConfig
from HABApp.config.models.mqtt import MqttConfig
from HABApp.config.models.openhab import OpenhabConfig
from HABApp.config.models.habapp import HABAppConfig
from HABApp.config.models.directories import DirectoriesConfig
from pydantic import Field


class ApplicationConfig(AppBaseModel):
    """Structure that contains the complete configuration"""

    directories: DirectoriesConfig = Field(default_factory=DirectoriesConfig)
    location: LocationConfig = Field(default_factory=LocationConfig)
    mqtt: MqttConfig = Field(default_factory=MqttConfig)
    openhab: OpenhabConfig = Field(default_factory=OpenhabConfig)
    habapp: HABAppConfig = Field(default_factory=HABAppConfig, in_file=False)



from __future__ import annotations

import asyncio
from typing import Final, TypeVar

import HABApp
from HABApp.core.connections import BaseConnection
from HABApp.core.connections._definitions import connection_log


T = TypeVar('T', bound=BaseConnection)


class ConnectionManager:
    def __init__(self):
        self.connections: dict[str, BaseConnection] = {}

    def add(self, connection: T) -> T:
        assert connection.name not in self.connections
        self.connections[connection.name] = connection
        connection_log.debug(f'Added {connection.name:s}')

        return connection

    def get(self, name: str) -> BaseConnection:
        return self.connections[name]

    def remove(self, name):
        con = self.get(name)
        if not con.is_shutdown:
            raise ValueError()
        self.connections.pop(name)

    async def on_application_shutdown(self):
        for c in self.connections.values():
            c.on_application_shutdown()

        tasks = [t.advance_status_task.wait() for t in self.connections.values()]
        await asyncio.gather(*tasks)

    def application_startup_complete(self):
        for c in self.connections.values():
            with HABApp.core.wrapper.ExceptionToHABApp(logger=c.log):
                c.application_startup_complete()

    def __repr__(self):
        return f'<{self.__class__.__name__}>'


connection_manager: Final = ConnectionManager()



from asyncio import Task, current_task, CancelledError
from typing import Callable, Awaitable, Any, Final, Optional

from HABApp.core.const import loop

_TASK_REFS = set()


class SingleTask:
    def __init__(self, coro: Callable[[], Awaitable[Any]], name: Optional[str] = None):
        if name is None:
            name = f'{self.__class__.__name__}_{coro.__name__}'

        self.coro: Final = coro
        self.name: Final = name
        self.task: Optional[Task] = None

    @property
    def is_running(self) -> bool:
        return self.task is not None

    def cancel(self) -> Optional[Task]:
        if (task := self.task) is None:
            return None

        self.task = None

        # we need the reference only when we cancel a task, otherwise the task is saved in the attribute
        _TASK_REFS.add(task)
        task.add_done_callback(_TASK_REFS.discard)

        task.cancel()
        return task

    async def cancel_wait(self):
        if task := self.cancel():
            try:
                await task
            except CancelledError:
                pass

    async def wait(self):
        if self.task is None:
            return None

        try:
            await self.task
        except CancelledError:
            pass

    def start(self) -> Task:
        self.cancel()
        self.task = task = loop.create_task(self._task_wrap(), name=self.name)
        return task

    def start_if_not_running(self) -> Task:
        if (task := self.task) is not None:
            return task

        self.task = task = loop.create_task(self._task_wrap(), name=self.name)
        return task

    async def _task_wrap(self):
        task = current_task(loop)

        # don't use try-finally because
        try:
            await self.coro()
        finally:
            # This also runs on asyncio.CancelledError so we have to make sure it's the same task
            if self.task is task:
                self.task = None



import logging
import logging.config
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from pathlib import Path
from typing import Any, Dict, List, Optional

import HABApp
from HABApp.config.config import CONFIG
from HABApp.config.errors import AbsolutePathExpected
from easyconfig.yaml import yaml_safe as _yaml_safe
from .buffered_logger import BufferedLogger
from .queue_handler import HABAppQueueHandler, SimpleQueue


def remove_memory_handler_from_cfg(cfg: dict, log: BufferedLogger):
    # find memory handlers
    m_handlers = {}
    for handler, handler_cfg in cfg.get('handlers', {}).items():
        if handler_cfg.get('class', '') == 'logging.handlers.MemoryHandler':
            log.error(f'"logging.handlers.MemoryHandler" is no longer required. Please remove from config ({handler})!')
            if 'target' in handler_cfg:
                m_handlers[handler] = handler_cfg['target']

    # remove them from config
    for h_name in m_handlers:
        cfg['handlers'].pop(h_name)
        log.warning(f'Removed {h_name:s} from handlers')

    # replace handlers in logger with target
    for logger_name, logger_cfg in cfg.get('loggers', {}).items():
        logger_handlers = logger_cfg.get('handlers', [])
        for i, logger_handler in enumerate(logger_handlers):
            replacement_handler = m_handlers.get(logger_handler)
            if replacement_handler is None:
                continue
            log.warning(f'Replaced {logger_handler} with {replacement_handler} for logger {logger_name}')
            logger_handlers[i] = replacement_handler


def get_logging_dict(path: Path, log: BufferedLogger) -> Optional[dict]:
    # config gets created on startup - if it gets deleted we do nothing here
    if not path.is_file():
        return None

    with path.open('r', encoding='utf-8') as file:
        cfg: Dict[str, Any] = _yaml_safe.load(file)

    # fix filenames
    for handler, handler_cfg in cfg.get('handlers', {}).items():
        # migrate handler
        if handler_cfg.get('class', '-') == 'HABApp.core.lib.handler.MidnightRotatingFileHandler':
            dst = 'HABApp.config.logging.MidnightRotatingFileHandler'
            handler_cfg['class'] = dst
            log.warning(f'Replaced class for handler "{handler:s}" with {dst}')

        if 'filename' not in handler_cfg:
            continue

        # fix encoding for FileHandlers - we always log utf-8
        if 'file' in handler_cfg.get('class', '').lower():
            enc = handler_cfg.get('encoding', '')
            if enc != 'utf-8':
                handler_cfg['encoding'] = 'utf-8'

        # make Filenames absolute path in the log folder if not specified
        p = Path(handler_cfg['filename'])
        if not p.is_absolute():
            # Our log folder ist not yet converted to path -> it is not loaded
            if not CONFIG.directories.logging.is_absolute():
                raise AbsolutePathExpected()

            # Use defined parent folder
            p = (CONFIG.directories.logging / p).resolve()
            handler_cfg['filename'] = str(p)

    # remove memory handlers
    remove_memory_handler_from_cfg(cfg, log)

    # make file version optional for config file
    if 'version' not in cfg:
        cfg['version'] = 1
    else:
        log.warning('Entry "version" is no longer required in the logging configuration file')

    # Allow the user to set his own logging levels (with aliases)
    for level, alias in cfg.pop('levels', {}).items():
        if not isinstance(level, int):
            level = logging._nameToLevel[level]
        logging.addLevelName(level, str(alias))
        log.debug(f'Added custom Level "{str(alias)}" ({level})')

    return cfg


def rotate_files():
    for wr in logging._handlerList:
        handler = wr()  # weakref -> call it to get object

        # only rotate these types
        if not isinstance(handler, (RotatingFileHandler, TimedRotatingFileHandler)):
            continue

        # Rotate only if files have content
        logfile = Path(handler.baseFilename)
        if not logfile.is_file() or logfile.stat().st_size <= 10:
            continue

        try:
            handler.acquire()
            handler.flush()
            handler.doRollover()
        except Exception as e:
            HABApp.core.wrapper.process_exception(rotate_files, e)
        finally:
            handler.release()


def inject_log_buffer(cfg: dict, log: BufferedLogger):
    from HABApp.core.const.log import TOPIC_EVENTS

    handler_cfg = cfg.setdefault('handlers', {})

    prefix = 'HABAppQueue_'

    # Check that the prefix is unique
    for handler_name in handler_cfg:
        if handler_name.startswith(prefix):
            raise ValueError(f'Handler may not start with {prefix:s}')

    # replace the event logs with the buffered one
    buffered_handlers = {}
    for log_name, log_cfg in cfg.get('loggers', {}).items():
        if not log_name.startswith(TOPIC_EVENTS):
            continue
        _handlers = {n: f'{prefix}{n}' for n in log_cfg['handlers']}
        buffered_handlers.update(_handlers)
        log_cfg['handlers'] = list(_handlers.values())

        # ensure propagate is disabled
        if log_cfg.get('propagate', False):
            log.warning(f'Propagate can not be set for {log_name}!')
        log_cfg['propagate'] = False

    if not buffered_handlers:
        return []

    handler_cfg = cfg.setdefault('handlers', {})
    q_handlers: List[HABAppQueueHandler] = []

    for handler_name, buffered_handler_name in buffered_handlers.items():
        q: SimpleQueue = SimpleQueue()
        handler_cfg[buffered_handler_name] = {'class': 'logging.handlers.QueueHandler', 'queue': q}

        qh = HABAppQueueHandler(q, handler_name, f'LogBuffer{handler_name:s}')
        q_handlers.append(qh)

    return q_handlers



from __future__ import annotations

from ._definitions import ConnectionStatus


class StatusTransitions:
    def __init__(self):
        self.status = ConnectionStatus.STARTUP
        self.manual: ConnectionStatus | None = None

        # Flags
        self.error = False
        self.setup = False
        self.shutdown = False

    def advance_status(self) -> ConnectionStatus | None:
        if self.manual is not None:
            self.status = status = self.manual
            self.manual = None
        else:
            if (status := self._next_step()) is None:
                return None
            self.status = status

        return status

    def is_connecting_or_connected(self):
        return self.status in (ConnectionStatus.CONNECTING, ConnectionStatus.CONNECTED, ConnectionStatus.ONLINE)

    def _set_manual(self, status: ConnectionStatus):
        assert self.manual is None
        self.manual = status

    def from_setup_to_disabled(self):
        assert self.status == ConnectionStatus.SETUP
        self._set_manual(ConnectionStatus.DISABLED)

    def from_connected_to_disconnected(self):
        assert self.status == ConnectionStatus.CONNECTED
        self._set_manual(ConnectionStatus.DISCONNECTED)

    def _next_step(self) -> ConnectionStatus:
        status = self.status

        if self.error:
            if self.is_connecting_or_connected():
                return ConnectionStatus.DISCONNECTED
            if status == ConnectionStatus.SETUP:
                return ConnectionStatus.DISABLED

        if self.setup:
            if self.is_connecting_or_connected():
                return ConnectionStatus.DISCONNECTED
            if status in (ConnectionStatus.STARTUP, ConnectionStatus.OFFLINE, ConnectionStatus.DISABLED):
                self.setup = False
                return ConnectionStatus.SETUP

        if self.shutdown:
            if self.is_connecting_or_connected():
                return ConnectionStatus.DISCONNECTED
            if status in (ConnectionStatus.STARTUP, ConnectionStatus.OFFLINE, ConnectionStatus.DISABLED):
                return ConnectionStatus.SHUTDOWN

        # Automatically reconnect if there are no errors
        if not self.error and status is ConnectionStatus.OFFLINE:
            return ConnectionStatus.CONNECTING

        # automatic transitions if no flags are set
        transitions = {
            ConnectionStatus.CONNECTING: ConnectionStatus.CONNECTED,
            ConnectionStatus.CONNECTED: ConnectionStatus.ONLINE,

            ConnectionStatus.DISCONNECTED: ConnectionStatus.OFFLINE,

            ConnectionStatus.SETUP: ConnectionStatus.CONNECTING,
        }
        return transitions.get(status)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.status} ' \
               f'[{"x" if self.error else " "}] Error, ' \
               f'[{"x" if self.setup else " "}] Setup>'

    def __eq__(self, other: ConnectionStatus):
        if not isinstance(other, ConnectionStatus):
            return NotImplemented
        return self.status == other



from typing import Type, TypeVar, Optional

from pendulum import UTC, DateTime
from pendulum import now as pd_now

from HABApp.core.internals import HINT_EVENT_FILTER_OBJ, HINT_EVENT_BUS_LISTENER
from HABApp.core.internals import uses_get_item, uses_item_registry, get_current_context
from HABApp.core.internals.item_registry import ItemRegistryItem
from HABApp.core.lib.parameters import TH_POSITIVE_TIME_DIFF, get_positive_time_diff
from eascheduler.const import local_tz
from .base_item_times import ChangedTime, ItemNoChangeWatch, ItemNoUpdateWatch, UpdatedTime
from .tmp_data import add_tmp_data as _add_tmp_data
from .tmp_data import restore_tmp_data as _restore_tmp_data
from ..const.hints import HINT_EVENT_CALLBACK

get_item = uses_get_item()
item_registry = uses_item_registry()


class BaseItem(ItemRegistryItem):
    """BaseItem, all items must inherit from this class
    """

    @classmethod
    def get_item(cls, name: str):
        """Returns an already existing item. If it does not exist or has a different item type an exception will occur.

        :param name: Name of the item
        """
        assert isinstance(name, str), type(name)
        item = get_item(name)
        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str):
        super().__init__(name)

        _now = pd_now(UTC)
        self._last_change: ChangedTime = ChangedTime(self._name, _now)
        self._last_update: UpdatedTime = UpdatedTime(self._name, _now)

    @property
    def last_change(self) -> DateTime:
        """
        :return: Timestamp of the last time when the item has been changed (read only)
        """
        return self._last_change.dt.in_timezone(local_tz).naive()

    @property
    def last_update(self) -> DateTime:
        """
        :return: Timestamp of the last time when the item has been updated (read only)
        """
        return self._last_update.dt.in_timezone(local_tz).naive()

    def __repr__(self):
        ret = ''
        for k in ['name', 'last_change', 'last_update']:
            ret += f'{", " if ret else ""}{k}: {getattr(self, k)}'
        return f'<{self.__class__.__name__} {ret:s}>'

    def watch_change(self, secs: TH_POSITIVE_TIME_DIFF) -> ItemNoChangeWatch:
        """Generate an event if the item does not change for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        secs = get_positive_time_diff(secs, round_digits=1)
        return self._last_change.add_watch(secs)

    def watch_update(self, secs: TH_POSITIVE_TIME_DIFF) -> ItemNoUpdateWatch:
        """Generate an event if the item does not receive and update for a certain period of time.
        Has to be called from inside a rule function.

        :param secs: secs after which the event will occur, max 1 decimal digit for floats
        :return: The watch obj which can be used to cancel the watch
        """
        secs = get_positive_time_diff(secs, round_digits=1)
        return self._last_update.add_watch(secs)

    def listen_event(self, callback: HINT_EVENT_CALLBACK,
                     event_filter: Optional[HINT_EVENT_FILTER_OBJ] = None) -> HINT_EVENT_BUS_LISTENER:
        """
        Register an event listener which listens to all event that the item receives

        :param callback: callback that accepts one parameter which will contain the event
        :param event_filter: Event filter. This is typically :class:`~HABApp.core.events.ValueUpdateEventFilter` or
            :class:`~HABApp.core.events.ValueChangeEventFilter` which will also trigger on changes/update from openhab
            or mqtt. Additionally it can be an instance of :class:`~HABApp.core.events.EventFilter` which additionally
            filters on the values of the event. It is also possible to group filters logically with, e.g.
            :class:`~HABApp.core.events.AndFilterGroup` and :class:`~HABApp.core.events.OrFilterGroup`
        """
        return get_current_context().rule.listen_event(self._name, callback=callback, event_filter=event_filter)

    def _on_item_added(self):
        """This function gets automatically called when the item is added to the item registry
        """
        _restore_tmp_data(self)

    def _on_item_removed(self):
        """This function gets automatically called when the item is removed from the item registry
        """
        _add_tmp_data(self)


# Hints for functions that use an item class as an input parameter
HINT_ITEM_OBJ = TypeVar('HINT_ITEM_OBJ', bound=BaseItem)
HINT_TYPE_ITEM_OBJ = Type[HINT_ITEM_OBJ]



import logging
import threading
import typing
from asyncio import sleep
from pathlib import Path

import HABApp
import HABApp.__cmd_args__ as cmd_args
from HABApp.core.files.errors import AlreadyHandledFileError
from HABApp.core.files.file import HABAppFile
from HABApp.core.files.folders import add_folder as add_habapp_folder
from HABApp.core.files.watcher import AggregatingAsyncEventHandler
from HABApp.core.logger import log_warning
from HABApp.core.wrapper import log_exception
from HABApp.runtime import shutdown
from .rule_file import RuleFile
from ..core.internals import uses_item_registry
from HABApp.core.internals.wrapped_function import run_function
from HABApp.core.connections import Connections

log = logging.getLogger('HABApp.Rules')

item_registry = uses_item_registry()


class RuleManager:

    def __init__(self, parent):
        assert isinstance(parent, HABApp.runtime.Runtime)
        self.runtime = parent

        self.files: typing.Dict[str, RuleFile] = {}

        # serialize loading
        self.__load_lock = threading.Lock()
        self.__files_lock = threading.Lock()

        # Processing
        self.__process_last_sec = 60

        self.watcher: typing.Optional[AggregatingAsyncEventHandler] = None

    async def setup(self):

        # shutdown
        shutdown.register_func(self.shutdown, msg='Cancel rule schedulers')

        if cmd_args.DO_BENCH:
            from HABApp.rule_manager.benchmark import BenchFile
            self.files['bench'] = file = BenchFile(self)
            ok = await run_function(file.load)
            if not ok:
                log.error('Failed to load Benchmark!')
                HABApp.runtime.shutdown.request_shutdown()
                return None
            file.check_all_rules()
            return

        class HABAppRuleFile(HABAppFile):
            LOGGER = log
            LOAD_FUNC = self.request_file_load
            UNLOAD_FUNC = self.request_file_unload

        path = HABApp.CONFIG.directories.rules
        folder = add_habapp_folder('rules/', path, 0)
        folder.add_file_type(HABAppRuleFile)
        self.watcher = folder.add_watch('.py', True)

        # Initial loading of rules
        HABApp.core.internals.wrap_func(self.load_rules_on_startup, logger=log).run()

    async def load_rules_on_startup(self):

        if HABApp.CONFIG.openhab.general.wait_for_openhab:
            c = Connections.get('openhab')
            while not (c.is_shutdown or c.is_disabled or c.is_online):
                await sleep(1)
        else:
            await sleep(1)

        # if we want to shut down we don't load the rules
        if HABApp.runtime.shutdown.requested:
            return None

        # trigger event for every file
        await self.watcher.trigger_all()
        return None

    @log_exception
    def get_rule(self, rule_name):
        found = []
        for file in self.files.values():
            if rule_name is None:
                for rule in file.rules.values():
                    found.append(rule)
            else:
                if rule_name in file.rules:
                    found.append(file.rules[rule_name])

        # if we want all return them
        if rule_name is None:
            return found

        # if we want a special one throw error
        if not found:
            raise KeyError(f'No Rule with name "{rule_name}" found!')
        return found if len(found) > 1 else found[0]

    async def request_file_unload(self, name: str, path: Path, request_lock=True):
        path_str = str(path)

        try:
            if request_lock:
                self.__load_lock.acquire()

            # Only unload already loaded files
            with self.__files_lock:
                already_loaded = path_str in self.files
            if not already_loaded:
                log_warning(log, f'Rule file {path} is not yet loaded and therefore can not be unloaded')
                return None

            log.debug(f'Removing file: {name}')
            with self.__files_lock:
                rule = self.files.pop(path_str)

            await run_function(rule.unload)
        finally:
            if request_lock:
                self.__load_lock.release()

    async def request_file_load(self, name: str, path: Path):
        path_str = str(path)

        # if we want to shut down we don't load the rules
        if HABApp.runtime.shutdown.requested:
            log.debug(f'Skip load of {name:s} because of shutdown')
            return None

        # Only load existing files
        if not path.is_file():
            log_warning(log, f'Rule file {name} ({path}) does not exist and can not be loaded!')
            return None

        with self.__load_lock:
            # Unload if we have already loaded
            with self.__files_lock:
                already_loaded = path_str in self.files
            if already_loaded:
                await self.request_file_unload(name, path, request_lock=False)

            log.debug(f'Loading file: {name}')
            with self.__files_lock:
                self.files[path_str] = file = RuleFile(self, name, path)

            ok = await run_function(file.load)
            if not ok:
                self.files.pop(path_str)
                log.warning(f'Failed to load {path_str}!')
                raise AlreadyHandledFileError()

        log.debug(f'File {name} successfully loaded!')

        # Do simple checks which prevent errors
        file.check_all_rules()

    def shutdown(self):
        for f in self.files.values():
            f.unload()



# noinspection PyProtectedMember
from sys import _getframe as sys_get_frame
from types import FrameType
from typing import Optional, Union, TYPE_CHECKING

from HABApp.core.errors import ContextNotSetError, ContextNotFoundError
from HABApp.core.internals.context import ContextProvidingObj, ContextBoundObj, Context

if TYPE_CHECKING:
    import HABApp


# noinspection PyProtectedMember
def get_current_context(obj: Optional[ContextProvidingObj] = None) -> 'HABApp.rule_ctx.HABAppRuleContext':
    if obj is not None:
        return obj._habapp_ctx

    frame: Optional[FrameType] = sys_get_frame(1)

    while frame is not None:

        ctx_obj: Union[None, object, ContextProvidingObj] = frame.f_locals.get('self')
        if ctx_obj is not None and isinstance(ctx_obj, ContextProvidingObj):
            ctx = ctx_obj._habapp_ctx
            if ctx is None:
                raise ContextNotSetError()
            return ctx

        frame = frame.f_back

    raise ContextNotFoundError()


class AutoContextBoundObj(ContextBoundObj):
    def __init__(self, parent_ctx: Optional['Context'] = None, **kwargs):
        if parent_ctx is None:
            parent_ctx = get_current_context()
        super().__init__(parent_ctx=parent_ctx, **kwargs)



from typing import Optional, Final
from typing import get_type_hints as _get_type_hints
from inspect import isclass

from HABApp.core.const import MISSING
from HABApp.core.const.hints import HINT_ANY_CLASS
from HABApp.core.internals import EventFilterBase


class EventFilter(EventFilterBase):
    """Triggers on event types and optionally on their values, too"""

    def __init__(self, event_class: HINT_ANY_CLASS, **kwargs):
        assert len(kwargs) < 3, 'EventFilter only allows up to two args that will be used to filter'
        assert isclass(event_class), f'Class for event required! Passed {event_class} ({type(event_class)})'

        self.event_class: Final = event_class

        # Property filters
        self.attr_name1: Optional[str] = None
        self.attr_value1 = None
        self.attr_name2: Optional[str] = None
        self.attr_value2 = None

        type_hints = _get_type_hints(event_class)

        for arg, value in kwargs.items():
            if value is MISSING:
                continue

            if arg not in type_hints:
                raise AttributeError(f'Filter attribute "{arg}" does not exist for "{event_class.__name__}"')

            if self.attr_name1 is None:
                self.attr_name1 = arg
                self.attr_value1 = value
            elif self.attr_name2 is None:
                self.attr_name2 = arg
                self.attr_value2 = value
            else:
                raise ValueError('Not implemented for more than 2 values!')

    def trigger(self, event) -> bool:
        if not isinstance(event, self.event_class):
            return False

        # Property filter
        if self.attr_name1 is not None:
            if getattr(event, self.attr_name1, None) != self.attr_value1:
                return False

            if self.attr_name2 is not None:
                if getattr(event, self.attr_name2, None) != self.attr_value2:
                    return False

        return True

    def describe(self) -> str:

        values = ''
        if self.attr_name1 is not None:
            values += f', {self.attr_name1}={self.attr_value1}'
        if self.attr_name2 is not None:
            values += f', {self.attr_name2}={self.attr_value2}'

        return f'{self.__class__.__name__}(type={self.event_class.__name__}{values})'


class TypeBoundEventFilter(EventFilter):
    """Class to inherit from if the filter criteria always is a hardcoded instance check"""

    def describe(self) -> str:

        values = ''
        if self.attr_name1 is not None:
            values += f'{self.attr_name1}={self.attr_value1}'
        if self.attr_name2 is not None:
            values += f'{", " if values else ""}{self.attr_name2}={self.attr_value2}'

        return f'{self.__class__.__name__}({values})'



import sys
import time
from enum import Enum
from typing import Final


class _MissingType(Enum):
    MISSING = object()

    def __repr__(self):
        return '<Missing>'


MISSING: Final = _MissingType.MISSING
STARTUP: Final = time.monotonic()

# Python Versions for feature control
PYTHON_38: Final = sys.version_info >= (3, 8)
PYTHON_39: Final = sys.version_info >= (3, 9)
PYTHON_310: Final = sys.version_info >= (3, 10)
PYTHON_311: Final = sys.version_info >= (3, 11)
PYTHON_312: Final = sys.version_info >= (3, 12)


# In python 3.11 there were changes to MyEnum(str, Enum), so we have to use the StrEnum
# https://docs.python.org/3/library/enum.html#enum.StrEnum
if PYTHON_311:
    # noinspection PyUnresolvedReferences
    from enum import StrEnum
else:
    class StrEnum(str, Enum):
        pass



from __future__ import annotations

import logging
from pathlib import Path
from typing import Callable, Awaitable, Any

from HABApp.core.files.errors import CircularReferenceError, DependencyDoesNotExistError, AlreadyHandledFileError
from HABApp.core.files.file.properties import FileProperties
from HABApp.core.files.manager.files import FILES, file_state_changed
from HABApp.core.wrapper import process_exception
from . import FileState

log = logging.getLogger('HABApp.files')


class HABAppFile:
    LOGGER: logging.Logger
    LOAD_FUNC: Callable[[str, Path], Awaitable[Any]]
    UNLOAD_FUNC: Callable[[str, Path], Awaitable[Any]]

    def __init__(self, name: str, path: Path, properties: FileProperties):
        self.name: str = name
        self.path: Path = path

        self.state: FileState = FileState.PENDING
        self.properties: FileProperties = properties
        log.debug(f'{self.name} added')

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} state: {self.state}>'

    def set_state(self, new_state: FileState):
        if self.state is new_state:
            return None

        self.state = new_state
        log.debug(f'{self.name} changed to {self.state}')
        file_state_changed(self)

    def _check_circ_refs(self, stack, prop: str):
        c: list[str] = getattr(self.properties, prop)
        for f in c:
            _stack = stack + (f, )
            if f in stack:
                raise CircularReferenceError(_stack)

            next_file = FILES.get(f)
            if next_file is not None:
                next_file._check_circ_refs(_stack, prop)

    def _check_properties(self):
        # check dependencies
        mis = set(filter(lambda x: x not in FILES, self.properties.depends_on))
        if mis:
            one = len(mis) == 1
            msg = f'File {self.path} depends on file{"" if one else "s"} that ' \
                  f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}'

            raise DependencyDoesNotExistError(msg)

        # check reload
        mis = set(filter(lambda x: x not in FILES, self.properties.reloads_on))
        if mis:
            one = len(mis) == 1
            log.warning(f'File {self.path} reloads on file{"" if one else "s"} that '
                        f'do{"es" if one else ""}n\'t exist: {", ".join(sorted(mis))}')

    def check_properties(self, log_msg: bool = False):
        if self.state is not FileState.PENDING and self.state is not FileState.DEPENDENCIES_ERROR:
            return None

        try:
            self._check_properties()
        except DependencyDoesNotExistError as e:
            if log_msg:
                log.error(e.msg)
            return self.set_state(FileState.DEPENDENCIES_ERROR)

        try:
            # check for circular references
            self._check_circ_refs((self.name, ), 'depends_on')
            self._check_circ_refs((self.name, ), 'reloads_on')
        except CircularReferenceError as e:
            log.error(f'Circular reference: {" -> ".join(e.stack)}')
            return self.set_state(FileState.DEPENDENCIES_ERROR)

        # Check if we can already load it
        self.set_state(FileState.DEPENDENCIES_OK if not self.properties.depends_on else FileState.DEPENDENCIES_MISSING)

    def check_dependencies(self):
        if self.state is not FileState.DEPENDENCIES_MISSING:
            return None

        for name in self.properties.depends_on:
            f = FILES.get(name, None)
            if f is None:
                return None
            if f.state is not FileState.LOADED:
                return None

        self.set_state(FileState.DEPENDENCIES_OK)
        return None

    async def load(self):
        assert self.state is FileState.DEPENDENCIES_OK, self.state

        try:
            await self.__class__.LOAD_FUNC(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(self.__class__.LOAD_FUNC, e, logger=self.LOGGER)
            self.set_state(FileState.FAILED)
            return None

        self.set_state(FileState.LOADED)
        return None

    async def unload(self):
        try:
            await self.__class__.UNLOAD_FUNC(self.name, self.path)
        except Exception as e:
            if not isinstance(e, AlreadyHandledFileError):
                process_exception(self.__class__.UNLOAD_FUNC, e, logger=self.LOGGER)
            self.set_state(FileState.FAILED)
            return None

        self.set_state(FileState.REMOVED)
        return None

    def file_changed(self, file: HABAppFile):
        name = file.name
        if name in self.properties.reloads_on:
            self.set_state(FileState.PENDING)



from __future__ import annotations

import logging
from asyncio import sleep
from time import monotonic
from typing import Final

import HABApp.openhab.events
from HABApp.config import CONFIG
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry, uses_event_bus
from HABApp.core.lib import SingleTask
from HABApp.openhab.connection.connection import OpenhabConnection

PING_CONFIG: Final = CONFIG.openhab.ping

log = logging.getLogger('HABApp.openhab.items')
Items = uses_item_registry()
EventBus = uses_event_bus()


class PingPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self.task: Final = SingleTask(self.ping_worker, 'OhQueueWorker')

        self.sent_value: float | None = None
        self.next_value: float | None = None
        self.timestamp_sent: float | None = None

        self.listener: HABApp.core.internals.EventBusListener | None = None

    async def on_connected(self):
        if not PING_CONFIG.enabled:
            return None

        self.sent_value = None
        self.next_value = None
        self.timestamp_sent = None

        self.listener = HABApp.core.internals.EventBusListener(
            HABApp.config.CONFIG.openhab.ping.item,
            HABApp.core.internals.wrap_func(self.ping_received),
            HABApp.core.events.EventFilter(HABApp.openhab.events.ItemStateUpdatedEvent)
        )
        EventBus.add_listener(self.listener)

        self.task.start()

    async def on_disconnected(self):
        await self.task.cancel_wait()

        if self.listener is not None:
            self.listener.cancel()
            self.listener = None

    async def ping_received(self, event: HABApp.openhab.events.ItemStateEvent):
        value = event.value
        if value != self.sent_value:
            return None

        # If we are queued up it's possible that we receive multiple pings
        # Then we only take the first one
        if self.next_value is None:
            self.next_value = round((monotonic() - self.timestamp_sent) * 1000, 1)

    async def ping_worker(self):
        try:
            log.debug('Ping started')

            item_name = PING_CONFIG.item

            if not (send_ping := Items.item_exists(item_name)):
                log.warning(f'Number item "{item_name:s}" does not exist!')

            while True:
                self.sent_value = self.next_value
                self.next_value = None
                self.timestamp_sent = monotonic()

                if send_ping:
                    HABApp.openhab.interface_async.async_post_update(
                        item_name,
                        f'{self.sent_value:.1f}' if self.sent_value is not None else None
                    )
                else:
                    send_ping = Items.item_exists(item_name)

                await sleep(PING_CONFIG.interval)

        except Exception as e:
            self.plugin_connection.process_exception(e, self.ping_worker)
        finally:
            log.debug('Ping stopped')



from pathlib import Path
from typing import List, Iterable, TYPE_CHECKING
import operator as _operator

from HABApp.core.const import MISSING

if TYPE_CHECKING:
    import HABApp


def list_files(folder: Path, file_filter: 'HABApp.core.files.watcher.file_watcher.EventFilterBase',
               recursive: bool = False) -> List[Path]:
    # glob is much quicker than iter_dir()
    files = folder.glob('**/*' if recursive else '*')
    return sorted(filter(lambda x: file_filter.notify(str(x)), files), key=lambda x: x.relative_to(folder))


def sort_files(files: Iterable[Path]) -> List[Path]:
    return sorted(files)


CMP_OPS = {
    'lt': _operator.lt, 'lower_than': _operator.lt,
    'le': _operator.le, 'lower_equal': _operator.le,
    'eq': _operator.eq, 'equal': _operator.eq,
    'ne': _operator.ne, 'not_equal': _operator.ne,
    'gt': _operator.gt, 'greater_than': _operator.gt,
    'ge': _operator.ge, 'greater_equal': _operator.ge,

    'is_': _operator.is_,
    'is_not': _operator.is_not,
}


def compare(value, **kwargs) -> bool:

    for name, cmp_value in kwargs.items():
        if cmp_value is MISSING:
            continue
        if CMP_OPS[name](value, cmp_value):
            return True
    return False



from asyncio import Future as _Future
from asyncio import run_coroutine_threadsafe as _run_coroutine_threadsafe
from contextvars import ContextVar as _ContextVar, Token
from typing import Any as _Any, Callable, Final, Optional
from typing import Callable as _Callable
from typing import Coroutine as _Coroutine
from typing import Optional as _Optional
from typing import TypeVar as _TypeVar

from HABApp.core.const import loop
from HABApp.core.const.const import PYTHON_310

if PYTHON_310:
    from typing import ParamSpec as _ParamSpec
else:
    from typing_extensions import ParamSpec as _ParamSpec


async_context = _ContextVar('async_ctx')


class AsyncContext:
    def __init__(self, value: str):
        self.value: Final = value
        self.token: Optional[Token[str]] = None
        self.parent: Optional[AsyncContext] = None

    def __enter__(self):
        assert self.token is None, self
        self.parent = async_context.get(None)
        self.token = async_context.set(self.value)

    def __exit__(self, exc_type, exc_val, exc_tb):
        async_context.reset(self.token)

    def __repr__(self):
        parent: str = ''
        if self.parent:
            parent = f'{self.parent} -> '
        return f'<{self.__class__.__name__} {parent:s}{self.value:s}>'


class AsyncContextError(Exception):
    def __init__(self, func: _Callable) -> None:
        super().__init__()
        self.func: _Callable = func

    def __str__(self):
        return f'Function "{self.func.__name__}" may not be called from an async context!'


_tasks = set()


def create_task(coro: _Coroutine, name: _Optional[str] = None) -> _Future:
    # https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    if async_context.get(None) is None:
        f = _run_coroutine_threadsafe(coro, loop)
        f.add_done_callback(_tasks.discard)
        return f
    else:
        t = loop.create_task(coro, name=name)
        t.add_done_callback(_tasks.discard)
        return t


_CORO_RET = _TypeVar('_CORO_RET')


def run_coro_from_thread(coro: _Coroutine[_Any, _Any, _CORO_RET], calling: _Callable) -> _CORO_RET:
    # This function call is blocking, so it can't be called in the async context
    if async_context.get(None) is not None:
        raise AsyncContextError(calling)

    fut = _run_coroutine_threadsafe(coro, loop)
    return fut.result()


P = _ParamSpec('P')
T = _TypeVar('T')


def run_func_from_async(func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    # we already have an async context
    if async_context.get(None) is not None:
        return func(*args, **kwargs)

    future = _run_coroutine_threadsafe(_run_func_from_async_helper(func, *args, **kwargs), loop)
    return future
    # Doc build fails if we enable this
    # Todo: Fix the Rule Runner
    # return future.result()


async def _run_func_from_async_helper(func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    return func(*args, **kwargs)



import collections
import logging
import runpy
import typing
from pathlib import Path

import HABApp
from HABApp.rule.rule_hook import HABAppRuleHook
from HABApp.core.internals import get_current_context

log = logging.getLogger('HABApp.Rules')


class RuleFile:
    def __init__(self, rule_manager, name: str, path: Path):
        from .rule_manager import RuleManager

        assert isinstance(rule_manager, RuleManager)
        self.rule_manager = rule_manager

        self.name: str = name
        self.path: Path = path

        self.rules = {}  # type: typing.Dict[str, HABApp.Rule]

        self.class_ctr: typing.Dict[str, int] = collections.defaultdict(lambda: 1)

    def suggest_rule_name(self, obj: 'HABApp.Rule') -> str:

        # if there is already a name set we make no suggestion
        if getattr(obj, 'rule_name', '') != '':
            return obj.rule_name.replace('ü', 'ue').replace('ö', 'oe').replace('ä', 'ae')

        name = obj.__class__.__name__
        found = self.class_ctr[name]
        self.class_ctr[name] += 1

        return f'{name:s}.{found:d}' if found > 1 else f'{name:s}'

    def check_all_rules(self):
        for rule in self.rules.values():  # type: HABApp.Rule
            get_current_context(rule).check_rule()

    def unload(self):

        # If we don't have any rules we can not unload
        if not self.rules:
            return None

        # unload all registered callbacks
        for rule in self.rules.values():  # type: HABApp.Rule
            get_current_context(rule).unload_rule()

        log.debug(f'File {self.name} successfully unloaded!')
        return None

    def __process_tc(self, tb: list):
        tb.insert(0, f"Could not load {self.path}!")
        return [line.replace('<module>', self.path.name) for line in tb]

    def create_rules(self, created_rules: list):

        rule_hook = HABAppRuleHook(created_rules.append, self.suggest_rule_name, self.rule_manager.runtime, self)

        # It seems like python 3.8 doesn't allow path like objects any more:
        # https://github.com/spacemanspiff2007/HABApp/issues/111
        with rule_hook:
            runpy.run_path(str(self.path), run_name=str(self.path), init_globals=rule_hook.in_dict())

    def load(self) -> bool:

        created_rules: typing.List[HABApp.rule.Rule] = []

        ign = HABApp.core.wrapper.ExceptionToHABApp(logger=log)
        ign.proc_tb = self.__process_tc

        with ign:
            self.create_rules(created_rules)

        if ign.raised_exception:
            # unload all rule instances which might have already been created otherwise they might
            # still listen to events and do stuff
            for rule in created_rules:
                with ign:
                    get_current_context(rule).unload_rule()
            return False

        if not created_rules:
            log.warning(f'Found no instances of HABApp.Rule in {str(self.path)}')
            return True

        with ign:
            for rule in created_rules:
                # ensure that we have a rule name
                rule.rule_name = self.suggest_rule_name(rule)

                # rule name must be unique for every file
                if rule.rule_name in self.rules:
                    raise ValueError(f'Rule name must be unique!\n"{rule.rule_name}" is already used!')

                self.rules[rule.rule_name] = rule
                log.info(f'Added rule "{rule.rule_name}" from {self.name}')

        if ign.raised_exception:
            # unload all rule instances which might have already been created otherwise they might
            # still listen to events and do stuff
            for rule in created_rules:
                with ign:
                    get_current_context(rule).unload_rule()
            return False

        return True



import logging
import sys
from pathlib import Path
from typing import Optional

from easyconfig import BaseModel
from pydantic import Field, validator

from HABApp.config.platform_defaults import get_log_folder

log = logging.getLogger('HABApp.Config')


class DirectoriesConfig(BaseModel):
    """Configuration of directories that are used"""

    logging: Path = Field(get_log_folder(Path('log')), description='Folder where the logs will be written to')
    rules: Path = Field(Path('rules'), description='Folder from which the rule files will be loaded')

    # Optional Folders
    param: Optional[Path] = Field(Path('params'), description='Folder from which the parameter files will be loaded')
    config: Optional[Path] = Field(Path('config'), description='Folder from which configuration files '
                                                               '(e.g. for textual thing configuration) will be loaded')
    lib: Optional[Path] = Field(Path('lib'), description='Folder where additional libraries can be placed')

    @validator('*')
    def ensure_folder(cls, value: Optional[Path]):
        import HABApp.__cmd_args__

        if value is None:
            return value

        # only resolve if we have a path set
        if not value.is_absolute() and HABApp.__cmd_args__.CONFIG_FILE.name:
            value = HABApp.__cmd_args__.CONFIG_FILE.parent / value
            value = value.resolve()

        if value == HABApp.__cmd_args__.CONFIG_FILE:
            raise ValueError(
                f'Can not be the same as the path for the HABApp config! ({HABApp.__cmd_args__.CONFIG_FILE})'
            )

        return value

    def create_folders(self):

        # create folder structure if it does not exist
        if not self.rules.is_dir():
            self.rules.mkdir()
        if not self.logging.is_dir():
            self.logging.mkdir()

        if not self.param.is_dir():
            log.info(f'Parameters disabled! Folder {self.param} does not exist!')
            self.param = None

        if not self.config.is_dir():
            log.info(f'Manual thing configuration disabled! Folder {self.config} does not exist!')
            self.config = None

        # add path for user libraries
        if self.lib is not None and self.lib.is_dir():
            lib_path = str(self.lib)
            if lib_path not in sys.path:
                sys.path.insert(0, lib_path)
                log.debug(f'Added library folder "{lib_path}" to system path')



from HABApp.openhab.connection.handler.func_async import \
    async_get_root, async_get_system_info, async_get_uuid, \
    async_get_things, async_get_thing, async_set_thing_cfg, async_set_thing_enabled,\
    async_get_items, async_item_exists, async_get_item, async_remove_item, async_create_item, \
    async_set_metadata, async_remove_metadata


from HABApp.openhab.connection.plugins import async_send_command, async_post_update



from HABApp.openhab.connection.handler.func_sync import \
    get_thing, set_thing_enabled,\
    item_exists, get_item, remove_item, create_item, \
    set_metadata, remove_metadata, \
    get_persistence_services, get_persistence_data, set_persistence_data, \
    get_link, remove_link, create_link


from HABApp.openhab.connection.plugins import send_command, post_update



from __future__ import annotations

import time
from pathlib import Path
from typing import Any

import msgspec

import HABApp
import HABApp.openhab.events
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.files.file import HABAppFile
from HABApp.core.files.folders import add_folder as add_habapp_folder
from HABApp.core.files.watcher import AggregatingAsyncEventHandler
from HABApp.core.lib import PendingFuture
from HABApp.core.logger import log_warning, HABAppError
from HABApp.openhab.connection.connection import OpenhabConnection
from HABApp.openhab.connection.plugins.plugin_things.cfg_validator import validate_cfg, InvalidItemNameError
from HABApp.openhab.connection.plugins.plugin_things.filters import THING_ALIAS, CHANNEL_ALIAS
from HABApp.openhab.connection.plugins.plugin_things.filters import apply_filters, log_overview
from ._log import log
from .file_writer import ItemsFileWriter
from .item_worker import create_item, cleanup_items
from .thing_worker import update_thing_cfg


class DuplicateItemError(Exception):
    pass


class TextualThingConfigPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self):
        super().__init__()
        self.created_items: dict[str, set[str]] = {}
        self.do_cleanup = PendingFuture(self.clean_items, 120)

        self.watcher: AggregatingAsyncEventHandler | None = None

        self.cache_ts: float = 0.0
        self.cache_cfg: list[dict[str, Any]] = []

    async def on_setup(self):
        path = HABApp.CONFIG.directories.config
        if path is None:
            return None

        if self.watcher is not None:
            return None

        class HABAppThingConfigFile(HABAppFile):
            LOGGER = log
            LOAD_FUNC = self.file_load
            UNLOAD_FUNC = self.file_unload

        folder = add_habapp_folder('config/', path, 50)
        folder.add_file_type(HABAppThingConfigFile)
        self.watcher = folder.add_watch('.yml')

    async def file_unload(self, prefix: str, path: Path):
        return None

    async def on_connected(self):
        if self.watcher is None:
            return None

        await self.load_thing_data(always=True)
        await self.watcher.trigger_all()

    @HABApp.core.wrapper.ignore_exception
    async def clean_items(self):
        items = set()
        for s in self.created_items.values():
            items.update(s)
        await cleanup_items(items)

    async def load_thing_data(self, always: bool) -> list[dict[str, Any]]:
        if always or not self.cache_cfg or time.time() - self.cache_ts > 20:
            self.cache_cfg = [msgspec.to_builtins(k) for k in await HABApp.openhab.interface_async.async_get_things()]
            self.cache_ts = time.time()
        return self.cache_cfg

    async def file_load(self, name: str, path: Path):
        # we have to check the naming structure because we get file events for the whole folder
        _name = path.name.lower()
        if not _name.startswith('thing_') or not _name.endswith('.yml'):
            log.warning(f'Name for "{name}" does not start with "thing_" -> skip!')
            return None

        # only load if we don't supply the data
        data = await self.load_thing_data(always=False)

        # remove created items
        self.created_items.pop(path.name, None)
        created_items = self.created_items.setdefault(path.name, set())

        # shedule cleanup
        self.do_cleanup.reset()

        # output file
        items_file_path = path.with_suffix('.items')
        items_file_writer = ItemsFileWriter()

        # we also get events when the file gets deleted
        if not path.is_file():
            log.debug(f'File {path} does not exist -> skipping Thing configuration!')
            return None
        log.debug(f'Loading {name}!')

        # load the config file
        with path.open(mode='r', encoding='utf-8') as file:
            try:
                cfg = HABApp.core.const.yml.load(file)
            except Exception as e:
                HABAppError(log).add_exception(e).dump()
                return None

        # validate configuration
        cfg = validate_cfg(cfg, path.name)
        if cfg is None:
            return None

        # if one entry has test set we show an overview of all the things
        if any(map(lambda x: x.test, cfg)):
            log_overview(data, THING_ALIAS, 'Thing overview')

        # process each thing part in the cfg
        for cfg_entry in cfg:
            test: bool = cfg_entry.test
            things = list(apply_filters(cfg_entry.filter, data, test))

            # show a warning we found no Things
            if not things:
                log_warning(log, f'No things matched for {cfg_entry.filter}')
                continue

            # update thing configuration
            if cfg_entry.thing_config:
                await update_thing_cfg(cfg_entry.thing_config, things, test)

            try:
                # item creation for every thing
                create_items = {}
                shown_types = set()
                for thing in things:
                    thing_context = {k: thing.get(alias, '') for k, alias in THING_ALIAS.items()}

                    # create items without channel
                    for item_cfg in cfg_entry.get_items(thing_context):
                        name = item_cfg.name
                        if name in create_items:
                            raise DuplicateItemError(f'Duplicate item: {name}')
                        create_items[name] = item_cfg

                    # Channel overview, only if we have something configured
                    if test and cfg_entry.channels:
                        __thing_type = thing_context['thing_type']
                        if __thing_type not in shown_types:
                            shown_types.add(__thing_type)
                            log_overview(thing['channels'], CHANNEL_ALIAS, heading=f'Channels for {__thing_type}')

                    # do channel things
                    for channel_cfg in cfg_entry.channels:
                        channels = apply_filters(channel_cfg.filter, thing['channels'], test)
                        for channel in channels:
                            channel_context = {k: channel.get(alias, '') for k, alias in CHANNEL_ALIAS.items()}
                            channel_context.update(thing_context)

                            for item_cfg in channel_cfg.get_items(channel_context):
                                item_cfg.link = channel['uid']
                                name = item_cfg.name

                                if name in create_items:
                                    raise DuplicateItemError(f'Duplicate item: {name}')
                                create_items[name] = item_cfg

                    # newline only if we create logs
                    if test and (cfg_entry.create_items or cfg_entry.channels):
                        log.info('')
            except InvalidItemNameError as e:
                HABAppError(log).add_exception(e).dump()
                continue
            except DuplicateItemError as e:
                # Duplicates should never happen, the user clearly made a mistake, that's why we exit here
                HABAppError(log).add_exception(e).dump()
                return None

            # Create all items
            for item_cfg in create_items.values():
                created = await create_item(item_cfg, test)
                if created:
                    created_items.add(item_cfg.name)

            self.do_cleanup.reset()

            items_file_writer.add_items(create_items.values())
            self.cache_cfg = []

        items_file_writer.create_file(items_file_path)



from __future__ import annotations

from asyncio import Queue, QueueEmpty
from asyncio import sleep
from typing import Any
from typing import Final

from HABApp.core.asyncio import run_func_from_async
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import ItemRegistryItem
from HABApp.core.lib import SingleTask
from HABApp.core.logger import log_info, log_warning
from HABApp.openhab.connection.connection import OpenhabConnection
from HABApp.openhab.connection.handler import convert_to_oh_type, post, put


class OutgoingCommandsPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None):
        super().__init__(name)

        self.add: bool = False

        self.queue: Queue[tuple[str | ItemRegistryItem, Any, bool]] = Queue()
        self.task_worker: Final = SingleTask(self.queue_worker, 'OhQueueWorker')
        self.task_watcher: Final = SingleTask(self.queue_watcher, 'OhQueueWatcher')

    async def _clear_queue(self):
        try:
            while True:
                self.queue.get_nowait()
        except QueueEmpty:
            pass

    async def on_connected(self):
        self.add = True
        self.task_worker.start()
        self.task_watcher.start()

    async def on_disconnected(self):
        self.add = False
        await self.task_worker.cancel_wait()
        await self.task_watcher.cancel_wait()
        await self._clear_queue()

    async def queue_watcher(self):
        log = self.plugin_connection.log
        first_msg_at = 150

        upper = first_msg_at
        lower = -1
        last_info_at = first_msg_at // 2

        while True:
            await sleep(10)
            size = self.queue.qsize()

            # small log msg
            if size > upper:
                upper = size * 2
                lower = size // 2
                log_warning(log, f'{size} messages in queue')
            elif size < lower:
                upper = max(size / 2, first_msg_at)
                lower = size // 2
                if lower <= last_info_at:
                    lower = -1
                    log_info(log, 'queue OK')
                else:
                    log_info(log, f'{size} messages in queue')

    # noinspection PyProtectedMember
    async def queue_worker(self):

        queue: Final = self.queue
        to_str: Final = convert_to_oh_type

        while True:
            try:
                while True:
                    item, state, is_cmd = await queue.get()

                    if not isinstance(item, str):
                        item = item._name

                    if not isinstance(state, str):
                        state = to_str(state)

                    if is_cmd:
                        await post(f'/rest/items/{item:s}', data=state)
                    else:
                        await put(f'/rest/items/{item:s}/state', data=state)
            except Exception as e:
                self.plugin_connection.process_exception(e, 'Outgoing queue worker')

    def async_post_update(self, item: str | ItemRegistryItem, state: Any):
        if not self.add:
            return None
        self.queue.put_nowait((item, state, False))

    def async_send_command(self, item: str | ItemRegistryItem, state: Any):
        if not self.add:
            return None
        self.queue.put_nowait((item, state, True))


OUTGOING_PLUGIN: Final = OutgoingCommandsPlugin()
async_post_update: Final = OUTGOING_PLUGIN.async_post_update
async_send_command: Final = OUTGOING_PLUGIN.async_send_command


def post_update(item: str | ItemRegistryItem, state: Any):
    """
    Post an update to the item

    :param item: item name or item
    :param state: new item state
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)

    run_func_from_async(async_post_update, item, state)


def send_command(item: str | ItemRegistryItem, command: Any):
    """
    Send the specified command to the item

    :param item: item name or item
    :param command: command
    """
    assert isinstance(item, (str, ItemRegistryItem)), type(item)

    run_func_from_async(async_send_command, item, command)



from __future__ import annotations

import logging
from asyncio import sleep
from datetime import datetime

from immutables import Map

import HABApp.openhab.events
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.connection.handler import map_null_str
from HABApp.openhab.connection.handler.func_async import async_get_items, \
    async_get_all_items_state, async_get_things
from HABApp.openhab.definitions import QuantityValue
from HABApp.openhab.definitions.rest import ThingResp
from HABApp.openhab.item_to_reg import fresh_item_sync, add_to_registry, remove_from_registry, add_thing_to_registry, \
    remove_thing_from_registry

log = logging.getLogger('HABApp.openhab.items')
Items = uses_item_registry()


class LoadOpenhabItemsPlugin(BaseConnectionPlugin[OpenhabConnection]):

    async def on_connected(self, context: OpenhabContext):
        if not context.created_items and not context.created_things:
            await self.load_items(context)
            await self.load_things(context)
        else:
            # Sleep so the event handler is running
            await sleep(0.1)

            if context.created_items:
                while await self.sync_items(context):
                    pass
            if context.created_things:
                while await self.sync_things(context):
                    pass

    async def load_items(self, context: OpenhabContext):
        from HABApp.openhab.map_items import map_item
        OpenhabItem = HABApp.openhab.items.OpenhabItem

        log.debug('Requesting items')
        items = await async_get_items()
        items_len = len(items)
        log.debug(f'Got response with {items_len} items')

        fresh_item_sync()

        # add all items
        for item in items:
            new_item = map_item(
                item.name, item.type, map_null_str(item.state), item.label,
                frozenset(item.tags), frozenset(item.groups), item.metadata
            )

            # error
            if new_item is None:
                continue
            add_to_registry(new_item, True)

        # remove items which are no longer available
        ist = set(Items.get_item_names())
        soll = {item.name for item in items}
        for k in ist - soll:
            if isinstance(Items.get_item(k), OpenhabItem):
                remove_from_registry(k)

        log.info(f'Updated {items_len:d} Items')

        created_items: dict[str, tuple[OpenhabItem, datetime]] = {
            i.name: (i, i.last_update) for i in Items.get_items() if isinstance(i, OpenhabItem)
        }
        context.created_items.update(created_items)

    async def sync_items(self, context: OpenhabContext):
        log.debug('Starting item state sync')
        created_items = context.created_items

        items = await async_get_all_items_state()

        synced = 0
        for item in items:
            # if the item is still None it was not initialized during the start of the item event listener
            if (new_state := map_null_str(item.state)) is None:
                continue

            # UoM item handling
            if ':' in item.type:
                new_state, _ = QuantityValue.split_unit(new_state)

            existing_item, existing_item_update = created_items[item.name]

            # noinspection PyProtectedMember
            new_value = existing_item._state_from_oh_str(new_state)

            if existing_item.value != new_value and existing_item.last_update == existing_item_update:
                existing_item.value = new_value
                log.debug(f'Re-synced value of {item.name:s}')
                synced += 1

        log.debug('Item state sync complete')
        return synced

    async def load_things(self, context: OpenhabContext):
        Thing = HABApp.openhab.items.Thing

        # try to update things, too
        log.debug('Requesting things')

        things = await async_get_things()
        thing_count = len(things)
        log.debug(f'Got response with {thing_count:d} things')

        created_things = {}
        for thing in things:
            t = add_thing_to_registry(thing)
            created_things[t.name] = (t, t.last_update)

        context.created_items.update(created_things)

        # remove things which were deleted
        ist = set(Items.get_item_names())
        soll = {thing.uid for thing in things}
        for k in ist - soll:
            if isinstance(Items.get_item(k), Thing):
                remove_thing_from_registry(k)
        log.info(f'Updated {thing_count:d} Things')

    async def sync_things(self, context: OpenhabContext):
        log.debug('Starting Thing sync')
        created_things = context.created_things

        synced = 0

        for thing in await async_get_things():
            existing_thing, existing_datetime = created_things[thing.uid]

            if thing_changed(existing_thing, thing) and existing_thing.last_update != existing_datetime:
                existing_thing.status = thing.status.status
                existing_thing.status_description = thing.status.description
                existing_thing.status_detail = thing.status.detail if thing.status.detail else ''
                existing_thing.label = thing.label
                existing_thing.location = thing.location
                existing_thing.configuration = Map(thing.configuration)
                existing_thing.properties = Map(thing.properties)
                log.debug(f'Re-synced {existing_thing.name:s}')
                synced += 1

        return synced


def thing_changed(old: HABApp.openhab.items.Thing, new: ThingResp) -> bool:
    return old.status != new.status.status or \
        old.status_detail != new.status.detail or \
        old.status_description != ('' if not new.status.description else new.status.description) or \
        old.label != new.label or \
        old.location != new.configuration or \
        old.configuration != new.configuration or \
        old.properties != new.properties



from threading import Lock
from typing import Optional, Dict, Any, Tuple, List

from HABApp.core.const import MISSING
from HABApp.core.items import Item
from .mode_base import HINT_BASE_MODE, BaseMode

LOCK = Lock()


class MultiModeItem(Item):
    """Prioritizer :class:`~HABApp.core.items.Item`
    """

    @classmethod
    def get_create_item(cls, name: str, initial_value=None, default_value=MISSING) -> 'MultiModeItem':
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param initial_value: state the item will have if it gets created
        :param default_value: Default value that will be sent if no mode is active
        :return: The created or existing item
        """
        item = super().get_create_item(name, initial_value)  # type: MultiModeItem
        item._default_value = default_value
        return item

    def __init__(self, name: str, initial_value=None, default_value=MISSING):
        super().__init__(name=name, initial_value=initial_value)

        self.__values_by_prio: Dict[int, HINT_BASE_MODE] = {}
        self.__values_by_name: Dict[str, HINT_BASE_MODE] = {}

        self._default_value = default_value

    def __remove_mode(self, name: str) -> bool:

        # Check if the mode exists
        mode_remove = self.__values_by_name.pop(name, None)
        if mode_remove is None:
            return False

        # Remove parent mapping, so we at least crash if someone attempts to do something with the mode
        mode_remove.parent = None

        # remove priority entry, too
        for prio, mode in self.__values_by_prio.items():
            if mode is mode_remove:
                self.__values_by_prio.pop(prio)
                return True

        raise RuntimeError(f'Mode {name} is missing!')

    def __sort_modes(self):
        # sort by priority and make lower prio known to the mode
        modes = sorted(self.__values_by_prio.items())
        self.__values_by_prio.clear()

        lower_mode: Optional[HINT_BASE_MODE] = None
        for prio, mode in modes:
            self.__values_by_prio[prio] = mode
            mode._set_mode_lower_prio(lower_mode)
            lower_mode = mode

        return None

    def remove_mode(self, name: str) -> bool:
        """Remove mode if it exists

        :param name: name of the mode (case-insensitive)
        :return: True if something was removed, False if nothing was found
        """
        assert isinstance(name, str), type(name)

        with LOCK:
            found = self.__remove_mode(name.lower())
            self.__sort_modes()
            return found

    def add_mode(self, priority: int, mode: HINT_BASE_MODE) -> 'MultiModeItem':
        """Add a new mode to the item, if it already exists it will be overwritten

        :param priority: priority of the mode
        :param mode: instance of the MultiMode class
        """
        assert isinstance(priority, int), type(priority)
        assert isinstance(mode, BaseMode), type(mode)

        name = mode.name.lower()

        with LOCK:
            # remove old mode
            self.__remove_mode(name)

            # add new mode
            self.__values_by_prio[priority] = mode
            self.__values_by_name[name] = mode
            mode.parent = self

            # resort
            self.__sort_modes()
        return self

    def all_modes(self) -> List[Tuple[int, HINT_BASE_MODE]]:
        """Returns a sorted list containing tuples with the priority and the mode

        :return: List with priorities and modes
        """
        return list(self.__values_by_prio.items())

    def get_mode(self, name: str) -> HINT_BASE_MODE:
        """Returns a created mode

        :param name: name of the mode (case insensitive)
        :return: The requested MultiModeValue
        """
        try:
            return self.__values_by_name[name.lower()]
        except KeyError:
            raise KeyError(f'Unknown mode "{name}"! Available: {", ".join(self.__values_by_name.keys())}') from None

    def calculate_value(self) -> Any:
        """Recalculate the value. If the new value is not ``MISSING`` the calculated value will be set as the item
        state and the corresponding events will be generated.

        :return: new value
        """

        # recalculate value
        new_value = MISSING
        for priority, child in self.__values_by_prio.items():
            new_value = child.calculate_value(new_value)

        # if nothing is set try the default
        if new_value is MISSING:
            new_value = self._default_value

        if new_value is not MISSING:
            self.post_value(new_value)

        return new_value

    def _on_item_removed(self):
        for p, mode in self.all_modes():
            mode.cancel()

        super()._on_item_removed()



import logging
import threading
from typing import Any, TypeVar
from typing import Dict, List

from HABApp.core.events import ComplexEventValue, ValueChangeEvent
from .base_listener import EventBusBaseListener
from HABApp.core.const.log import TOPIC_EVENTS

event_log = logging.getLogger(TOPIC_EVENTS)
habapp_log = logging.getLogger('HABApp')

_TYPE_LISTENER = TypeVar('_TYPE_LISTENER', bound=EventBusBaseListener)


class EventBus:
    def __init__(self):
        self._lock = threading.Lock()
        self._listeners: Dict[str, List[EventBusBaseListener]] = {}

    def post_event(self, topic: str, event: Any):
        assert isinstance(topic, str), type(topic)

        if not isinstance(event, str):
            event_prv = str(event)
        else:
            event_prv = event[:120] + ' ...' if len(event) > 120 else event
            event_prv = "'" + event_prv.replace('\n', '\\n') + "'"

        event_log.info(f'{topic:>20s}: {event_prv}')

        # Sometimes we have nested data structures which we need to set the value.
        # Once the value in the item registry is updated the data structures provide no benefit thus
        # we unpack the corresponding value
        try:
            if isinstance(event.value, ComplexEventValue):
                event.value = event.value.value
            if isinstance(event, ValueChangeEvent) and isinstance(event.old_value, ComplexEventValue):
                event.old_value = event.old_value.value
        except AttributeError:
            pass

        # Notify all listeners
        listeners = self._listeners.get(topic, None)
        if listeners is not None:
            for listener in listeners:
                listener.notify_listeners(event)
        return None

    def add_listener(self, listener: _TYPE_LISTENER):
        assert isinstance(listener, EventBusBaseListener)
        assert isinstance(listener.topic, str) and listener.topic

        with self._lock:
            item_listeners = self._listeners.setdefault(listener.topic, [])

            # don't add the same listener twice
            if listener in item_listeners:
                habapp_log.warning(f'Event listener for {listener.describe()} has already been added!')
                return None

            # add listener
            item_listeners.append(listener)
            habapp_log.debug(f'Added event listener for {listener.describe()}')
            return None

    def remove_listener(self, listener: _TYPE_LISTENER):
        assert isinstance(listener, EventBusBaseListener)
        assert isinstance(listener.topic, str) and listener.topic

        with self._lock:
            item_listeners = self._listeners.get(listener.topic, [])

            # print warning if we try to remove it twice
            if listener not in item_listeners:
                habapp_log.warning(f'Event listener for {listener.describe()} has already been removed!')
                return None

            # remove listener
            item_listeners.remove(listener)
            habapp_log.debug(f'Removed event listener for {listener.describe()}')
            return None

    def remove_all_listeners(self):
        with self._lock:
            self._listeners.clear()



import argparse
import ctypes
import os
import sys
import time
import typing
from pathlib import Path

# Global var if we want to run the benchmark
DO_BENCH = False
DO_DEBUG = False
CONFIG_FILE = Path()


def get_uptime() -> float:

    if sys.platform == 'linux':
        with open('/proc/uptime') as f:
            return float(f.readline().split()[0])

    if sys.platform == 'win32':
        lib = ctypes.windll.kernel32
        lib.GetTickCount64.restype = ctypes.c_uint64
        return lib.GetTickCount64() / 1000

    raise NotImplementedError(f'Not supported on {sys.platform}')


def parse_args(passed_args=None) -> argparse.Namespace:
    global DO_BENCH, DO_DEBUG

    parser = argparse.ArgumentParser(description='Start HABApp')
    parser.add_argument(
        '-c',
        '--config',
        help='Path to configuration folder (where the config.yml is located)',
        default=None
    )
    parser.add_argument(
        '-wos',
        '--wait_os_uptime',
        help='Waits for the specified os uptime before starting HABApp',
        type=int,
        default=None
    )
    parser.add_argument(
        '-b',
        '--benchmark',
        help='Do a Benchmark based on the current config',
        action='store_true'
    )
    parser.add_argument(
        '-di',
        '--debug-info',
        help='Print debug information',
        action='store_true'
    )
    args = parser.parse_args(passed_args)

    DO_BENCH = args.benchmark
    DO_DEBUG = args.debug_info

    if args.wait_os_uptime:
        args.wait_os_uptime = max(0, args.wait_os_uptime)
        uptime = get_uptime()
        if uptime < args.wait_os_uptime:
            diff = args.wait_os_uptime - uptime
            print(f'Waiting {diff:.0f} seconds before starting HABApp ...', end='')
            time.sleep(diff)
            print(' done!')

    return args


def find_config_folder(arg_config_path: typing.Optional[Path]) -> Path:
    global CONFIG_FILE

    if arg_config_path is None:
        # Nothing is specified, we try to find the config automatically
        check_path = []
        try:
            working_dir = Path(os.getcwd())
            check_path.append( working_dir / 'HABApp')
            check_path.append( working_dir.with_name('HABApp'))
            check_path.append( working_dir.parent.with_name('HABApp'))
        except ValueError:
            # the ValueError gets raised if the working_dir or its parent is empty (e.g. c:\ or /)
            pass

        check_path.append(Path.home() / 'HABApp')   # User home

        # if we run in a venv check the venv, too
        v_env = os.environ.get('VIRTUAL_ENV', '')
        if v_env:
            check_path.append(Path(v_env) / 'HABApp')  # Virtual env dir
    else:
        arg_config_path = Path(arg_config_path).resolve()

        # in case the user specifies the config.yml we automatically switch to the parent folder
        if arg_config_path.name.lower() == 'config.yml' and arg_config_path.is_file():
            arg_config_path = arg_config_path.parent

        # Override automatic config detection if something is specified through command line
        check_path = [arg_config_path]

    for config_folder in check_path:
        config_folder = config_folder.resolve()
        if not config_folder.is_dir():
            continue

        config_file = config_folder / 'config.yml'
        if config_file.is_file():
            CONFIG_FILE = config_file
            return config_folder

    # we have specified a folder, but the config does not exist so we will create it
    if arg_config_path is not None and arg_config_path.is_dir():
        return arg_config_path

    # we have nothing found and nothing specified -> exit
    print('Config file "config.yml" not found!')
    print('Checked folders:\n - ' + '\n - '.join(str(k) for k in check_path))
    print('Please create file or specify a folder with the "-c" arg switch.')
    sys.exit(1)



from typing import Any, Optional, Mapping

import aiohttp

import HABApp
from HABApp.core.const.json import dump_json


CLIENT: Optional[aiohttp.ClientSession] = None


async def create_client():
    global CLIENT
    assert CLIENT is None

    CLIENT = aiohttp.ClientSession(json_serialize=dump_json, loop=HABApp.core.const.loop)

    from HABApp.runtime import shutdown
    shutdown.register_func(CLIENT.close, msg='Closing generic http connection')


def get(url: str, params: Optional[Mapping[str, str]] = None, **kwargs: Any)\
        -> aiohttp.client._RequestContextManager:
    """http get request

    :param url: Request URL
    :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                   to be sent as parameters in the query string of the new request.
                   `Params example
                   <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
    :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                   for further possible kwargs
    :return: awaitable
    """
    assert CLIENT is not None
    return CLIENT.get(url, params=params, **kwargs)


def post(url: str, params: Optional[Mapping[str, str]] = None,
         data: Any = None, json: Any = None, **kwargs: Any) -> aiohttp.client._RequestContextManager:
    """http post request

    :param url: Request URL
    :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                   to be sent as parameters in the query string of the new request.
                   `Params example
                   <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
    :param data: Dictionary, bytes, or file-like object to send in the body of the request
                 (optional)
    :param json: Any json compatible python object, json and data parameters could not be used at the same time.
                 (optional)
    :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                   for further possible kwargs
    :return: awaitable
    """
    assert CLIENT is not None
    return CLIENT.post(url, params=params, data=data, json=json, **kwargs)


def put(url: str, params: Optional[Mapping[str, str]] = None,
        data: Any = None, json: Any = None, **kwargs: Any) -> aiohttp.client._RequestContextManager:
    """http put request

    :param url: Request URL
    :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                   to be sent as parameters in the query string of the new request.
                   `Params example
                   <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
    :param data: Dictionary, bytes, or file-like object to send in the body of the request
                 (optional)
    :param json: Any json compatible python object, json and data parameters could not be used at the same time.
                 (optional)
    :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                   for further possible kwargs
    :return: awaitable
    """
    assert CLIENT is not None
    return CLIENT.put(url, params=params, data=data, json=json, **kwargs)


def delete(url: str, params: Optional[Mapping[str, str]] = None, **kwargs: Any)\
        -> aiohttp.client._RequestContextManager:
    """http delete request

    :param url: Request URL
    :param params: Mapping, iterable of tuple of key/value pairs (e.g. dict)
                   to be sent as parameters in the query string of the new request.
                   `Params example
                   <https://docs.aiohttp.org/en/stable/client_quickstart.html#passing-parameters-in-urls>`_
    :param kwargs: See `aiohttp request <https://docs.aiohttp.org/en/stable/client_reference.html#aiohttp.request>`_
                   for further possible kwargs
    :return: awaitable
    """
    assert CLIENT is not None
    return CLIENT.delete(url, params=params, **kwargs)


def get_client_session() -> aiohttp.ClientSession:
    """Return the aiohttp
    `client session object <https://docs.aiohttp.org/en/stable/client_reference.html#client-session>`_
    for use in aiohttp libraries

    :return: session object
    """
    assert CLIENT is not None
    return CLIENT



from __future__ import annotations

from typing import Any, Iterable

import HABApp
from HABApp.config import CONFIG
from HABApp.config.models.mqtt import QOS
from HABApp.core.asyncio import run_func_from_async
from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import uses_post_event, uses_get_item, uses_item_registry
from HABApp.core.lib import SingleTask
from HABApp.core.wrapper import process_exception
from HABApp.mqtt.connection.connection import MqttPlugin
from HABApp.mqtt.events import MqttValueChangeEvent, MqttValueUpdateEvent
from HABApp.mqtt.mqtt_payload import get_msg_payload

SUBSCRIBE_CFG = CONFIG.mqtt.subscribe


class SubscriptionHandler(MqttPlugin):
    def __init__(self):
        super().__init__(task_name='MqttSubscribe')
        self.runtime_subs: dict[str, int] = {}
        self.subscribed_to: dict[str, int] = {}

        self.sub_task = SingleTask(self.apply_subscriptions, 'ApplySubscriptionsTask')

    async def interface_subscribe(self, topic_or_topics: str | Iterable[tuple[str, int | None]],
                                  qos: QOS | None = None):
        """
        Subscribe to a MQTT topic. Note that subscriptions made this way are volatile and will only remain until
        the next restart.

        :param topic_or_topics: MQTT topic or multiple topic qos pairs to subscribe to
        :param qos: QoS, can be 0, 1 or 2.  If not specified value from configuration file will be used.
        """

        if qos is None:
            qos = SUBSCRIBE_CFG.qos
        if not isinstance(topic_or_topics, str):
            for _t, _q in topic_or_topics:
                self.runtime_subs[_t] = _q if _q is not None else qos
        else:
            self.runtime_subs[topic_or_topics] = qos

        if self.plugin_connection.context is not None:
            await self.apply_subscriptions()

    async def interface_unsubscribe(self, topic_or_topics: str | Iterable[str]):
        """
        Unsubscribe from a MQTT topic

        :param topic_or_topics: MQTT topic
        """

        if isinstance(topic_or_topics, str):
            topic_or_topics = [topic_or_topics]

        for topic in topic_or_topics:
            self.runtime_subs.pop(topic, None)

        if self.plugin_connection.context is not None:
            await self.apply_subscriptions()

    def subscription_cfg_changed(self):
        if not self.plugin_connection.is_online:
            return None
        self.sub_task.start_if_not_running()

    async def unsubscribe(self, topics: list[str] | None):
        log = self.plugin_connection.log

        if (client := self.plugin_connection.context) is None:
            return None

        if topics is None:
            topics = sorted(set(self.runtime_subs) | set(self.subscribed_to))

        if not topics:
            return None

        log.debug('Unsubscribing from:')
        for topic in topics:
            log.debug(f' - "{topic:s}"')

        await client.unsubscribe(topics)

        for topic in topics:
            self.subscribed_to.pop(topic)

    async def apply_subscriptions(self):
        log = self.plugin_connection.log
        default_qos = SUBSCRIBE_CFG.qos

        client = self.plugin_connection.context
        assert client is not None

        target: dict[str, int] = {}

        # If our connection has errors we'll do a disconnect cycle anyway, so we don't even try to subscribe to anything
        # Unsubscribing has the corresponding handling, so we call that every time
        if not self.plugin_connection.has_errors:
            # subscription from config
            for topic, qos in CONFIG.mqtt.subscribe.topics:
                target[topic] = qos if qos is not None else default_qos
            # runtime subscriptions overwrite the subscriptions from the config file
            for topic, qos in self.runtime_subs.items():
                target[topic] = qos

        unsubscribe = []
        for sub_topic, sub_qos in sorted(self.subscribed_to.items()):
            if sub_topic not in target or target[sub_topic] != sub_qos:
                unsubscribe.append(sub_topic)

        await self.unsubscribe(unsubscribe)

        if subscribe := [(topic, qos) for topic, qos in target.items() if topic not in self.subscribed_to]:
            log.debug('Subscribing to:')
            for topic, qos in subscribe:
                log.debug(f' - "{topic}" (QoS {qos:d})')

            await client.subscribe(subscribe)

            for topic, qos in subscribe:
                self.subscribed_to[topic] = qos

        log.debug('Subscriptions successfully updated')

    async def on_connected(self):
        await super().on_connected()
        self.sub_task.start_if_not_running()
        await self.sub_task.wait()

    async def on_disconnected(self):
        await super().on_disconnected()
        await self.sub_task.cancel_wait()

        # without errors, it's a graceful disconnect
        if not self.plugin_connection.has_errors:
            await self.unsubscribe(None)

    async def mqtt_task(self):
        client = self.plugin_connection.context
        assert client is not None

        async with client.messages() as messages:
            async for message in messages:

                try:
                    topic, payload = get_msg_payload(message)
                    if topic is None:
                        continue

                    msg_to_event(topic, payload, message.retain)
                except Exception as e:
                    process_exception('mqtt payload handling', e, logger=self.plugin_connection.log)


post_event = uses_post_event()
get_item = uses_get_item()
Items = uses_item_registry()


def msg_to_event(topic: str, payload: Any, retain: bool):

    _item = None    # type: HABApp.mqtt.items.MqttBaseItem | None
    try:
        _item = get_item(topic)   # type: HABApp.mqtt.items.MqttBaseItem
    except ItemNotFoundException:
        # only create items for if the message has the retain flag
        if retain:
            _item = Items.add_item(HABApp.mqtt.items.MqttItem(topic))

    # we don't have an item -> we process only the event
    if _item is None:
        post_event(topic, MqttValueUpdateEvent(topic, payload))
        return None

    # Remember state and update item before doing callbacks
    _old_state = _item.value
    _item.set_value(payload)

    post_event(topic, MqttValueUpdateEvent(topic, payload))
    if payload != _old_state:
        post_event(topic, MqttValueChangeEvent(topic, payload, _old_state))


SUBSCRIPTION_HANDLER = SubscriptionHandler()


async_subscribe = SUBSCRIPTION_HANDLER.interface_subscribe
async_unsubscribe = SUBSCRIPTION_HANDLER.interface_unsubscribe


def subscribe(topic_or_topics: str | Iterable[tuple[str, int | None]], qos: QOS | None = None):
    """
    Subscribe to a MQTT topic. Note that subscriptions made this way are volatile and will only remain until
    the next restart.

    :param topic_or_topics: MQTT topic or multiple topic qos pairs to subscribe to
    :param qos: QoS, can be 0, 1 or 2.  If not specified value from configuration file will be used.
    """
    run_func_from_async(async_subscribe(topic_or_topics, qos))


def unsubscribe(topic_or_topics: str | Iterable[str]):
    """
    Unsubscribe from a MQTT topic

    :param topic_or_topics: MQTT topic
    """
    run_func_from_async(async_subscribe(topic_or_topics))



from __future__ import annotations

from typing import Final, TYPE_CHECKING, Generic, TypeVar, Callable, Awaitable, Any

from HABApp.core.lib import SingleTask

if TYPE_CHECKING:
    from .plugin_callback import PluginCallbackHandler
    from .base_connection import BaseConnection

T = TypeVar('T', bound='BaseConnection')


class BaseConnectionPlugin(Generic[T]):
    def __init__(self, name: str | None = None):
        super().__init__()

        if name is None:
            name = self.__class__.__name__
            if name[-6:].lower() == 'plugin':
                name = name[:-6]

        self.plugin_connection: T = None
        self.plugin_name: Final = name
        self.plugin_callbacks: dict[str, PluginCallbackHandler] = {}

    def on_application_shutdown(self):
        pass


class BaseConnectionPluginConnectedTask(BaseConnectionPlugin[T]):
    def __init__(self, task_coro: Callable[[], Awaitable[Any]],
                 task_name: str, name: str | None = None):
        super().__init__(name)
        self.task: Final = SingleTask(task_coro, name=task_name)

    async def on_connected(self):
        self.task.start()

    async def on_disconnected(self):
        await self.task.cancel_wait()



import logging
import logging.config
from pathlib import Path
from typing import List

import pydantic

import HABApp
import eascheduler
from HABApp import __version__
from HABApp.config.config import CONFIG
from HABApp.config.logging import HABAppQueueHandler
from .errors import InvalidConfigError, AbsolutePathExpected
from .logging import create_default_logfile, get_logging_dict, rotate_files, inject_log_buffer
from .logging.buffered_logger import BufferedLogger

log = logging.getLogger('HABApp.Config')


def load_config(config_folder: Path):

    CONFIG.set_file_path(config_folder / 'config.yml')

    logging_cfg_path = config_folder / 'logging.yml'
    create_default_logfile(logging_cfg_path)

    loaded_logging = False

    # Try load the logging config
    try:
        load_logging_cfg(logging_cfg_path)
        loaded_logging = True
    except (AbsolutePathExpected, InvalidConfigError):
        pass

    load_habapp_cfg(do_print=not loaded_logging)

    if not loaded_logging:
        load_logging_cfg(logging_cfg_path)

    # Watch folders, so we can reload the config on the fly
    filter = HABApp.core.files.watcher.FileEndingFilter('.yml')
    watcher = HABApp.core.files.watcher.AggregatingAsyncEventHandler(
        config_folder, config_files_changed, filter, watch_subfolders=False
    )
    HABApp.core.files.watcher.add_folder_watch(watcher)

    HABApp.runtime.shutdown.register_func(stop_queue_handlers, last=True, msg='Stopping logging threads')
    CONFIG.habapp.logging.subscribe_for_changes(set_flush_delay)


def set_flush_delay():
    HABAppQueueHandler.FLUSH_DELAY = CONFIG.habapp.logging.flush_every


async def config_files_changed(paths: List[Path]):
    for path in paths:
        if path.name == 'config.yml':
            load_habapp_cfg()
        if path.name == 'logging.yml':
            load_logging_cfg(path)


def load_habapp_cfg(do_print=False):
    try:
        CONFIG.load_config_file()
    except pydantic.ValidationError as e:
        for line in str(e).splitlines():
            if do_print:
                print(line)
            else:
                log.error(line)
        raise InvalidConfigError from None

    # check if folders exist and print warnings, maybe because of missing permissions
    if not CONFIG.directories.rules.is_dir():
        log.warning(f'Folder for rules files does not exist: {CONFIG.directories.rules}')

    CONFIG.directories.create_folders()

    log.debug(f'Local Timezone: {eascheduler.const.local_tz}')
    location = CONFIG.location
    eascheduler.set_location(location.latitude, location.longitude, location.elevation)

    log.debug('Loaded HABApp config')


QUEUE_HANDLER: List['HABAppQueueHandler'] = []


def stop_queue_handlers():
    for qh in QUEUE_HANDLER:
        qh.signal_stop()
    while QUEUE_HANDLER:
        qh = QUEUE_HANDLER.pop()
        qh.stop()


def load_logging_cfg(path: Path):
    # stop buffered handlers
    stop_queue_handlers()

    buf_log = BufferedLogger()
    cfg = get_logging_dict(path, buf_log)

    if CONFIG.habapp.logging.use_buffer:
        q_handlers = inject_log_buffer(cfg, buf_log)
    else:
        q_handlers = []

    # load prepared logging
    try:
        logging.config.dictConfig(cfg)
    except Exception as e:
        print(f'Error loading logging config: {e}')
        log.error(f'Error loading logging config: {e}')
        raise InvalidConfigError from None

    rotate_files()

    # start buffered handlers
    for qh in q_handlers:
        QUEUE_HANDLER.append(qh)
        qh.start()

    logging.getLogger('HABApp').info(f'HABApp Version {__version__}')

    # write buffered messages
    buf_log.flush(log)
    return None



import logging
import time
from asyncio import Future, create_task, sleep
from pathlib import Path
from typing import Optional

import HABApp
from HABApp.core.files.file import FileState
from HABApp.core.files.folders import get_prefixes
from . import FILES

log = logging.getLogger('HABApp.files')


TASK: Optional[Future] = None
TASK_SLEEP: float = 0.3
TASK_DURATION: float = 15


async def process_file(name: str, file: Path):
    global TASK

    # unload file
    if not file.is_file():
        existing = FILES.pop(name, None)
        if existing is None:
            return None

        await existing.unload()
        log.debug(f'Removed {existing.name}')
        return None

    # add file
    FILES[name] = HABApp.core.files.file.create_file(name, file)
    if TASK is None:
        TASK = create_task(_process())


async def _process():
    global TASK

    prefixes = get_prefixes()

    ct = -1
    log_msg = False
    last_process = time.time()

    try:
        while True:

            # wait until files are stable
            while ct != len(FILES):
                ct = len(FILES)
                await sleep(TASK_SLEEP)

            # check files for dependencies etc.
            for file in FILES.values():
                file.check_properties(log_msg)

            # Load order
            for prefix in prefixes:
                file_loaded = False
                for name in filter(lambda x: x.startswith(prefix), sorted(FILES.keys())):
                    file = FILES[name]
                    file.check_dependencies()

                    if file.state is FileState.DEPENDENCIES_OK:
                        await file.load()
                        last_process = time.time()
                        file_loaded = True
                        break
                if file_loaded:
                    break

            # if we don't have any files left to load we sleep!
            if not any(map(lambda x: x.state is FileState.DEPENDENCIES_OK, FILES.values())):
                await sleep(TASK_SLEEP)

            # Emit an error message during the last run
            if log_msg:
                break
            log_msg = time.time() - last_process > TASK_DURATION

    except Exception as e:
        HABApp.core.wrapper.process_exception('file load worker', e, logger=log)
    finally:
        TASK = None
    log.debug('Worker done!')



from ._definitions import ConnectionStatus, CONNECTION_HANDLER_NAME

# isort: split

from .base_plugin import BaseConnectionPlugin
from .plugin_callback import PluginCallbackHandler
from .base_connection import BaseConnection

# isort: split

from .manager import connection_manager as Connections

# isort: split

from .plugins import ConnectionStateToEventBusPlugin, AutoReconnectPlugin



from HABApp.__debug_info__ import get_debug_info


def test_debug_info():
    info = get_debug_info()

    for line in info.splitlines():
        if ':' in line:
            assert not line.strip().endswith(':')
